#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool libray. It defines the base classes
for"""

__all__ = ["CONTROLLER_TEMPLATE", "CTRL_TYPE_MAP", "TYPE_MAP", "TYPE_MAP_OBJ",
           "TypeData", "DTYPE_MAP", "DACCESS_MAP", "DataInfo",
           "ControllerLibrary", "ControllerClass"]

__docformat__ = 'restructuredtext'

import types
import inspect

from taurus.core.util.containers import CaselessDict

from sardana import DataType, DataFormat, DataAccess, \
    to_dtype_dformat, to_daccess, \
    ElementType, TYPE_ELEMENTS, InvalidId
from sardana.sardanameta import SardanaLibrary, SardanaClass
from sardana.pool.poolmotor import PoolMotor
from sardana.pool.poolpseudomotor import PoolPseudoMotor
from sardana.pool.poolmotorgroup import PoolMotorGroup
from sardana.pool.poolmeasurementgroup import PoolMeasurementGroup
from sardana.pool.poolcountertimer import PoolCounterTimer
from sardana.pool.poolzerodexpchannel import Pool0DExpChannel
from sardana.pool.poolonedexpchannel import Pool1DExpChannel
from sardana.pool.pooltwodexpchannel import Pool2DExpChannel
from sardana.pool.poolpseudocounter import PoolPseudoCounter
from sardana.pool.poolinstrument import PoolInstrument
from sardana.pool.poolioregister import PoolIORegister
from sardana.pool.poolcontroller import PoolController, \
    PoolPseudoMotorController, PoolPseudoCounterController
from sardana.pool.controller import Controller, MotorController, \
    CounterTimerController, ZeroDController, OneDController, TwoDController, \
    PseudoMotorController, PseudoCounterController, IORegisterController
from sardana.pool.controller import Type, Access, Description, DefaultValue, \
    FGet, FSet, Memorize, Memorized, MaxDimSize


#: String containing template code for a controller class
CONTROLLER_TEMPLATE = """class @controller_name@(@controller_type@):
    \"\"\"@controller_name@ description.\"\"\"

"""

ET = ElementType

#: a dictionary dict<:data:`~sardana.ElementType`, class>
#: mapping element type enumeration with the corresponding controller pool class
#: (:class:`~sardana.pool.poolcontroller.PoolController` or sub-class of it).
CTRL_TYPE_MAP = {
    ET.Motor          : PoolController,
    ET.CTExpChannel   : PoolController,
    ET.OneDExpChannel : PoolController,
    ET.TwoDExpChannel : PoolController,
    ET.PseudoMotor    : PoolPseudoMotorController,
    ET.PseudoCounter  : PoolPseudoCounterController,
    ET.IORegister     : PoolController,
}

#: dictionary dict<:data:`~sardana.ElementType`, :class:`tuple`>
#: where tuple is a sequence:
#:
#: #. type string representation
#: #. family
#: #. internal pool class
#: #. automatic full name
#: #. controller class
TYPE_MAP = {
    ET.Controller       : ("Controller", "Controller", CTRL_TYPE_MAP, "controller/{klass}/{name}", Controller),
    ET.Instrument       : ("Instrument", "Instrument", PoolInstrument, "{full_name}", None),
    ET.Motor            : ("Motor", "Motor", PoolMotor, "motor/{ctrl_name}/{axis}", MotorController),
    ET.CTExpChannel     : ("CTExpChannel", "ExpChannel", PoolCounterTimer, "expchan/{ctrl_name}/{axis}", CounterTimerController),
    ET.ZeroDExpChannel  : ("ZeroDExpChannel", "ExpChannel", Pool0DExpChannel, "expchan/{ctrl_name}/{axis}", ZeroDController),
    ET.OneDExpChannel   : ("OneDExpChannel", "ExpChannel", Pool1DExpChannel, "expchan/{ctrl_name}/{axis}", OneDController),
    ET.TwoDExpChannel   : ("TwoDExpChannel", "ExpChannel", Pool2DExpChannel, "expchan/{ctrl_name}/{axis}", TwoDController),
    ET.PseudoMotor      : ("PseudoMotor", "Motor", PoolPseudoMotor, "pm/{ctrl_name}/{axis}", PseudoMotorController),
    ET.PseudoCounter    : ("PseudoCounter", "ExpChannel", PoolPseudoCounter, "pc/{ctrl_name}/{axis}", PseudoCounterController),
    ET.MotorGroup       : ("MotorGroup", "MotorGroup", PoolMotorGroup, "mg/{pool_name}/{name}", None),
    ET.MeasurementGroup : ("MeasurementGroup", "MeasurementGroup", PoolMeasurementGroup, "mntgrp/{pool_name}/{name}", None),
    ET.IORegister       : ("IORegister", "IORegister"      , PoolIORegister, "ioregister/{ctrl_name}/{axis}", IORegisterController),
}

class TypeData(object):
    """Information for a specific Element type"""

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

#: dictionary
#: dict<:data:`~sardana.ElementType`, :class:`~sardana.pool.poolmetacontroller.TypeData`>
TYPE_MAP_OBJ = {}
for t, d in TYPE_MAP.items():
    o = TypeData(type=t, name=d[0], family=d[1], klass=d[2] ,
                 auto_full_name=d[3], ctrl_klass=d[4])
    TYPE_MAP_OBJ[t] = o


class ControllerLibrary(SardanaLibrary):
    """Object representing a python module containning controller classes.
    Public members:

        - module - reference to python module
        - f_path - complete (absolute) path and filename
        - f_name - filename (including file extension)
        - path - complete (absolute) path
        - name - module name (without file extension)
        - controller_list - list<ControllerClass>
        - exc_info - exception information if an error occured when loading
                     the module
    """

    def __init__(self, **kwargs):
        kwargs['manager'] = kwargs.pop('pool')
        kwargs['elem_type'] = ElementType.ControllerLibrary
        SardanaLibrary.__init__(self, **kwargs)

    add_controller = SardanaLibrary.add_meta_class
    get_controller = SardanaLibrary.get_meta_class
    get_controllers = SardanaLibrary.get_meta_classes
    has_controller = SardanaLibrary.has_meta_class

    def serialize(self, *args, **kwargs):
        kwargs = SardanaLibrary.serialize(self, *args, **kwargs)
        kwargs['pool'] = self.get_manager().name
        kwargs['id'] = InvalidId
        return kwargs

    @property
    def controllers(self):
        return self.meta_classes


class DataInfo(object):

    def __init__(self, name, dtype, dformat=DataFormat.Scalar,
                 access=DataAccess.ReadWrite, description="",
                 default_value=None, memorized='true',
                 fget=None, fset=None, maxdimsize=None):
        self.name = name
        self.dtype = dtype
        self.dformat = dformat
        self.access = access
        self.description = description
        self.default_value = default_value
        self.memorized = memorized
        self.fget = fget or "get%s" % name
        self.fset = fset or "set%s" % name
        if maxdimsize == None:
            if dformat == DataFormat.Scalar:
                maxdimsize = ()
            elif dformat == DataFormat.OneD:
                maxdimsize = 2048,
            elif dformat == DataFormat.TwoD:
                maxdimsize = 2048, 2048
        self.maxdimsize = maxdimsize

    def copy(self):
        s = self
        d = DataInfo(s.name, s.dtype, dformat=s.dformat, access=s.access,
                     description=s.description, default_value=s.default_value,
                     memorized=s.memorized, fget=s.fget, fset=s.fset,
                     maxdimsize=self.maxdimsize)
        return d

    @classmethod
    def toDataInfo(klass, name, info):
        info = CaselessDict(info)
        dtype = info[Type]
        dtype, dformat = to_dtype_dformat(dtype)
        default_value = info.get(DefaultValue)
        description = info.get(Description, '')
        daccess = info.get(Access, DataAccess.ReadWrite)
        daccess = to_daccess(daccess)
        memorized = info.get(Memorize, Memorized)
        maxdimsize = info.get(MaxDimSize)
        fget = info.get(FGet)
        fset = info.get(FSet)
        if default_value is not None and dtype != DataType.String:
            if type(default_value) in types.StringTypes:
                default_value = eval(default_value)
        return DataInfo(name, dtype, dformat=dformat, access=daccess,
                        description=description, default_value=default_value,
                        memorized=memorized, fget=fget, fset=fset,
                        maxdimsize=maxdimsize)

    def toDict(self):
        return { 'name' : self.name, 'type' : DataType.whatis(self.dtype),
                 'format' : DataFormat.whatis(self.dformat),
                 'access' : DataAccess.whatis(self.access),
                 'description' : self.description,
                 'default_value' : self.default_value,
                 'memorized' : self.memorized,
                 'maxdimsize' : self.maxdimsize }

    def serialize(self, *args, **kwargs):
        kwargs.update(self.toDict())
        return kwargs

    def __repr__(self):
        return "{0}(name={1}, type={2}, format={3}, access={4})".format(
            self.__class__.__name__, self.name, DataType[self.dtype],
            DataFormat[self.dformat], DataAccess[self.access])

#class PropertyInfo(DataInfo):

#    def __init__(self, name, dtype, dformat=DataFormat.Scalar,
#                 description="", default_value=None):
#        DataInfo.__init__(self, name, dtype, dformat, access=DataAcces.ReadWrite,
#                          description=description, default_value=default_value)


#class AttributeInfo(DataInfo):

#    def __init__(self, name, dtype, dformat=DataFormat.Scalar,
#                 access=DataAccess.ReadWrite, description=""):
#        DataInfo.__init__(self, name, dtype, dformat, access=DataAcces.ReadWrite,
#                          description=description, default_value=None)

class ControllerClass(SardanaClass):
    """Object representing a python controller class.
       Public members:

           - name - class name
           - klass - python class object
           - lib - ControllerLibrary object representing the module where the
             controller is."""

    def __init__(self, **kwargs):
        kwargs['manager'] = kwargs.pop('pool')
        kwargs['elem_type'] = ElementType.ControllerClass
        SardanaClass.__init__(self, **kwargs)

        self.types = []
        self.dict_extra = {}
        self.api_version = 1

        klass = self.klass
        # Generic controller information
        self.ctrl_features = tuple(klass.ctrl_features)

        self.ctrl_properties = props = CaselessDict()
        for k, v in klass.class_prop.items():  # old member
            props[k] = DataInfo.toDataInfo(k, v)
        for k, v in klass.ctrl_properties.items():
            props[k] = DataInfo.toDataInfo(k, v)

        self.ctrl_attributes = ctrl_attrs = CaselessDict()
        for k, v in klass.ctrl_attributes.items():
            ctrl_attrs[k] = DataInfo.toDataInfo(k, v)

        self.axis_attributes = axis_attrs = CaselessDict()
        for k, v in klass.ctrl_extra_attributes.items():  # old member
            axis_attrs[k] = DataInfo.toDataInfo(k, v)
        for k, v in klass.axis_attributes.items():
            axis_attrs[k] = DataInfo.toDataInfo(k, v)

        self.types = types = self.__build_types()
        self.type_names = map(ElementType.whatis, types)

        if ElementType.PseudoMotor in types:
            self.motor_roles = tuple(klass.motor_roles)
            self.pseudo_motor_roles = tuple(klass.pseudo_motor_roles)
            if len(self.pseudo_motor_roles) == 0:
                self.pseudo_motor_roles = (klass.__name__,)
            self.dict_extra['motor_roles'] = self.motor_roles
            self.dict_extra['pseudo_motor_roles'] = self.pseudo_motor_roles

        if ElementType.PseudoCounter in types:
            self.counter_roles = tuple(klass.counter_roles)
            self.pseudo_counter_roles = tuple(klass.pseudo_counter_roles)
            if len(self.pseudo_counter_roles) == 0:
                self.pseudo_counter_roles = (klass.__name__,)
            self.dict_extra['counter_roles'] = self.counter_roles
            self.dict_extra['pseudo_counter_roles'] = self.pseudo_counter_roles

        if ElementType.IORegister in types:
            self.dict_extra['predefined_values'] = klass.predefined_values

        init_args = inspect.getargspec(klass.__init__)
        if init_args.varargs is None or init_args.keywords is None:
            self.api_version = 0

    def __build_types(self):
        types = []
        klass = self.klass
        for _type, type_data in TYPE_MAP_OBJ.items():
            if not _type in TYPE_ELEMENTS:
                continue
            if issubclass(klass, type_data.ctrl_klass):
                types.append(_type)
        return types

    def serialize(self, *args, **kwargs):
        kwargs = SardanaClass.serialize(self, *args, **kwargs)
        kwargs['id'] = InvalidId
        kwargs['pool'] = self.get_manager().name
        kwargs['gender'] = self.gender
        kwargs['model'] = self.model
        kwargs['organization'] = self.organization
        kwargs['types'] = self.type_names
        if len(self.type_names):
            kwargs['main_type'] = self.type_names[0]
        else:
            kwargs['main_type'] = None
        kwargs['api_version'] = self.api_version
        kwargs.update(self.dict_extra)
        return kwargs

    @property
    def controller_class(self):
        return self.klass

    @property
    def gender(self):
        return self.klass.gender

    @property
    def model(self):
        return self.klass.model

    @property
    def organization(self):
        return self.klass.organization





#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains a definition for ViewOptions"""

from __future__ import with_statement
from __future__ import print_function

__all__ = ['ViewOption']


def ViewOptionMeta(name, bases, attrs):
    for k in attrs['_DEFAULT_VIEW_OPTIONS']:
        attrs[k] = k
    return type(name, bases, attrs)


class ViewOption(object):
    __metaclass__ = ViewOptionMeta

    _DEFAULT_VIEW_OPTIONS = {
        'ShowDial' : True,
    }

    @classmethod
    def init_options(cls, d):
        d.update(cls._DEFAULT_VIEW_OPTIONS)
        return d

    @classmethod
    def reset_option(cls, d, name):
        if name in cls._DEFAULT_VIEW_OPTIONS:
            d[name] = cls._DEFAULT_VIEW_OPTIONS[name]
        else:
            del d[name]





#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains a taurus text editor widget."""

__all__ = ["SardanaEditor"]

__docformat__ = 'restructuredtext'

import os
import os.path as osp
import tempfile
import shutil
import functools

from taurus.external.qt import Qt

from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.editor import TaurusBaseEditor
from taurus.qt.qtgui.util import ActionFactory
from taurus.qt.qtgui.dialog import ProtectTaurusMessageBox
from macrotree import MacroSelectionDialog
from elementtree import SardanaElementTreeWidget

from sardana.taurus.qt.qtcore.tango.sardana.model import SardanaBaseProxyModel, \
    SardanaElementTypeModel, SardanaTypeTreeItem, SardanaRootTreeItem

from sardanabasewizard import SardanaBaseWizard, SardanaBasePage

_MACRO_LIB_TEMPLATE = """#!/usr/bin/env python
{copyright}
# don't forget to place every new macro here!
__all__ = []

__docformat__ = 'restructuredtext'

{non_sardana_imports}
from sardana.macroserver.macro import macro, Macro, Type
{sardana_imports}
# Place your code here!

"""

_MACRO_CLASS_TEMPLATE = """

class {macro_name}(Macro):
    \"\"\"{macro_name} description.\"\"\"

    # uncomment the following lines as necessary. Otherwise you may delete them
    #param_def = []
    #result_def = []
    #hints = {}
    #env = (,)
    
    # uncomment the following lines if need prepare. Otherwise you may delete them
    #def prepare(self):
    #    pass
        
    def run(self):
        pass

"""

_MACRO_FUNCTION_TEMPLATE = """

@macro()
def {macro_name}(self):
    self.output("Running {macro_name}...")

"""

class NewElementWizard(SardanaBaseWizard):
    pass


class ChooseElementTypePage(SardanaBasePage):

    def __init__(self, parent=None):
        SardanaBasePage.__init__(self, parent)

        self.setTitle('Please select type of element you wish to create')
        layout = Qt.QVBoxLayout()
        self.setLayout(layout)


class SardanaLibProxyModel(SardanaBaseProxyModel):

    ALLOWED_TYPES = 'MacroLibrary',  #'ControllerLibrary',

    def filterAcceptsRow(self, sourceRow, sourceParent):
        sourceModel = self.sourceModel()
        idx = sourceModel.index(sourceRow, 0, sourceParent)
        treeItem = idx.internalPointer()

        if isinstance(treeItem, SardanaRootTreeItem):
            return True
        if isinstance(treeItem, SardanaTypeTreeItem):
            return treeItem.itemData() in self.ALLOWED_TYPES
        return True


class SardanaLibTreeWidget(SardanaElementTreeWidget):

    KnownPerspectives = { "Type" : {
                          "label" : "By lib",
                          "icon" : ":/python-package.png",
                          "tooltip" : "View elements by library",
                          "model" : [SardanaLibProxyModel, SardanaElementTypeModel],
                        },
    }
    DftPerspective = "Type"


class SardanaEditor(TaurusBaseEditor, TaurusBaseWidget):

    def __init__(self, parent=None, designMode=None):
        name = self.__class__.__name__
        self._base_tmp_dir = None
        self._tmp_dir = None
        self._is_filesystem_prepared = False
        self.call__init__wo_kw(TaurusBaseEditor, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)
        self._elementTree = elementTree = \
            SardanaLibTreeWidget(self, with_navigation_bar=False,
                                 with_filter_widget=False,)
        elementTree.treeView().setColumnHidden(1, True)
        self.connect(self._elementTree, Qt.SIGNAL("itemDoubleClicked"),
                     self.on_element_clicked)
        self.insertWidget(0, self._elementTree)
        self.setAutoTooltip(False)

    def setTemporaryDirectory(self, tmp_dir):
        self._base_tmp_dir = tmp_dir
        self._is_filesystem_prepared = False

    def getTemporaryDirectory(self):
        return self._base_tmp_dir

    def createMenuActions(self):
        af = ActionFactory()
        on_save = functools.partial(self.on_save, apply=False)
        on_save_apply = functools.partial(self.on_save, apply=True)

        self.new_action = af.createAction(self, "New...",
                icon='document-new', tip="Create a new macro or controller class",
                triggered=self.on_new, shortcut=Qt.QKeySequence.New)
        self.open_action = af.createAction(self, "Open...",
                icon='document-open', tip="Open macro(s) or controller(s)",
                triggered=self.on_open, shortcut=Qt.QKeySequence.Open)
        self.save_action = af.createAction(self, "Save",
                icon='document-save', tip="Save the current selected item",
                triggered=on_save)
        self.save_and_apply_action = af.createAction(self, "Save && apply",
                triggered=on_save_apply, icon='document-save',
                tip="Save the current selected item and apply the new code",
                shortcut=Qt.QKeySequence.Save)
        self.revert_action = af.createAction(self, "Revert",
                icon='edit-undo', tip="Revert the current selected item code",
                triggered=self.on_revert)

        io_actions = [self.new_action, self.open_action, self.save_action,
                      self.revert_action]
        actions = [self.new_action, self.open_action, self.save_action,
                   self.save_and_apply_action, self.revert_action]
        return actions, io_actions

    def register_editorstack(self, editorstack):
        TaurusBaseEditor.register_editorstack(self, editorstack)
        self.connect(editorstack, Qt.SIGNAL('refresh_save_all_action()'),
                     self.refresh_save_and_apply_action)

    def refresh_save_and_apply_action(self):
        self.save_and_apply_action.setEnabled(self.save_action.isEnabled())

    def on_element_clicked(self, item, item_column):
        item_data = item.itemData()
        interfaces = item_data.interfaces
        if 'MacroCode' in interfaces:
            self.open_macros([item_data])
        elif 'MacroLibrary' in interfaces:
            self.open_macro_libraries([item_data])

    @ProtectTaurusMessageBox(title="A error occured trying to create a class")
    def on_new(self):

        elem_types = "Macro function", "Macro class", "Macro library", \
            "Motor controller class", "Counter/Timer controller class", \
            "Pseudo motor controller class"

        msg = "Please select type of element you wish to create"
        elem_type, ok = Qt.QInputDialog.getItem(self, "New", msg, elem_types,
                                                current=0, editable=False)
        if not ok:
            return

        idx = elem_types.index(elem_type)
        if idx == 0:
            return self.new_macro_function()
        elif idx == 1:
            return self.new_macro_class()
        elif idx == 2:
            return self.new_macro_library()
        raise NotImplementedError("Sorry! Not implemented yet.")

    def new_macro(self, template):
        macro_server = self.getModelObj()

        msg = "Please select the library where you want to place the new macro"

        macro_libraries = macro_server.getElementsOfType("MacroLibrary")
        macro_lib_names = macro_libraries.keys()
        macro_lib_names.sort()
        macro_lib_name, ok = Qt.QInputDialog.getItem(self, "Macro library", msg,
            macro_lib_names, current=0, editable=False)
        if not ok:
            return
        macro_lib_name = str(macro_lib_name)
        macro_lib = macro_libraries[macro_lib_name]

        fname, path = macro_lib.file_path, macro_lib.path
        fname = fname[fname.index(osp.sep) + 1:]  # transform into relative path
        local_filename = osp.join(self._tmp_dir, fname)

        msg = "Please give new macro name"

        macros = macro_lib.elements
        valid = False
        while not valid:
            macro_name, ok = Qt.QInputDialog.getText(self, "Macro name", msg)
            if not ok:
                return
            if macro_name in macros:
                res = Qt.QMessageBox.information(self, "Macro already exists",
                    "A macro named '%s' already exists in '%s'.\n"
                    "Please give a different macro name"
                    % (macro_name, macro_lib_name),
                    Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel, Qt.QMessageBox.Ok)
                if res == Qt.QMessageBox.Cancel:
                    return
                continue
            valid = True

        pars = dict(macro_name=macro_name)
        code = template.format(**pars)

        editorstack = self.editorStack()
        idx = editorstack.has_filename(local_filename)
        if idx is None:
            if not self._prepare_path(path):
                raise Exception("Error trying to prepare path for %s", path)
            self.debug("Creating local file %s...", local_filename)
            fname, lib_code, line = macro_server.GetMacroCode((macro_lib_name,))
            fd = file(local_filename, "w")
            fd.write(lib_code)
            fd.close()
            self.debug("Loading local file %s...", local_filename)
            self.load(local_filename, goto=None)
        else:
            pass
        editorstack.set_current_filename(local_filename)
        editor = editorstack.get_current_editor()
        editor.set_cursor_position(editor.get_position("eof"))
        editor.append(code)
        return macro_lib, macro_name

    def new_macro_class(self):
        macro_info = self.new_macro(_MACRO_CLASS_TEMPLATE)
        if macro_info is None:
            return
        macro_lib, macro_name = macro_info

    def new_macro_function(self):
        macro_info = self.new_macro(_MACRO_FUNCTION_TEMPLATE)
        if macro_info is None:
            return
        macro_lib, macro_name = macro_info

    def new_macro_library(self):
        ms = self.getModelObj()
        ms_path = ms.getMacroPathObj()
        directory, ok = Qt.QInputDialog.getItem(self, "New macro module",
            "Select the directory where you want the new macro module to " \
            "be placed",
            ms_path.macro_path, current=0, editable=False)
        if not ok:
            return
        directory = str(directory)
        ok = 0
        while not ok:
            mod, ok = Qt.QInputDialog.getText(self, "New macro module",
                "Module name:", Qt.QLineEdit.Normal, "")
            if not ok:
                return
            mod = str(mod)
            m, ext = osp.splitext(mod)
            if len(ext):
                if ext != ".py" :
                    Qt.QMessageBox.critical(self, "Wrong extension",
                        "When given, file extension<b> MUST</b> be <code>.py</code>.")
                    ok = 0
                    continue
            else:
                mod = mod + ".py"
        self._prepare_path(directory)
        filename = osp.join(directory, mod)
        rel_filename = filename[filename.index(osp.sep) + 1:]  # transform into relative path
        local_filename = osp.join(self._tmp_dir, rel_filename)
        f = file(local_filename, "w")

        #TODO: ask for additional imports
        #TODO: check if door environment has copyright variable

        pars = dict(copyright="", non_sardana_imports="",
                    sardana_imports="")
        code = _MACRO_LIB_TEMPLATE.format(**pars)
        f.write(code)
        f.close()
        self.editorStack().load(local_filename)

    @ProtectTaurusMessageBox(title="An error occured trying to open a macro class")
    def on_open(self):
        ms = self.getModelObj()
        ms_tree = MacroSelectionDialog(self, model_name=ms.getNormalName())
        ms_tree.exec_()
        if ms_tree.result() != Qt.QDialog.Accepted:
            return
        self.open_macros(ms_tree.getSelectedMacros())

    @ProtectTaurusMessageBox(title="An error occured trying to open macro(s)")
    def open_macros(self, macros):
        editorstack = self.editorStack()

        all_any = Qt.QMessageBox.YesToAll, Qt.QMessageBox.NoToAll
        yes_any = Qt.QMessageBox.Yes, Qt.QMessageBox.YesToAll
        no_any = Qt.QMessageBox.No, Qt.QMessageBox.NoToAll

        last_answer = Qt.QMessageBox.No
        for macro_info in macros:
            name, fname = macro_info.name, macro_info.file_path
            module = macro_info.module
            fname = fname[fname.index(osp.sep) + 1:]  # transform into relative path
            local_filename = osp.join(self._tmp_dir, fname)

            idx = editorstack.has_filename(local_filename)
            if idx is not None and last_answer not in all_any:
                last_answer = Qt.QMessageBox.question(self,
                    "Macro file '{0}' already opened".format(module),
                    "All changes to <b>{0}</b> will be lost."
                    "<br>Do you want to revert file from the server ignoring "
                    "any possible changes you (may) have made?".format(module),
                    Qt.QMessageBox.Yes | Qt.QMessageBox.YesToAll | \
                    Qt.QMessageBox.No | Qt.QMessageBox.NoToAll,
                    Qt.QMessageBox.No)

            if idx is None or last_answer in yes_any:
                if not self._prepare_path(macro_info.path):
                    Qt.QMessageBox.warning(self,
                        "Error trying to prepare '{0}'".format(module),
                        "An error occured trying to prepare '{0}'".format(module),
                        Qt.QMessageBox.Ok, Qt.QMessageBox.Ok)
                    continue
                _, code, line = self.get_macro_code(module, name)
                line = int(line)
                self.debug("Creating local file %s...", local_filename)
                fd = file(local_filename, "w")
                fd.write(code)
                fd.close()
                if idx is None:
                    self.debug("Loading local file %s...", local_filename)
                    self.load(local_filename, goto=line)
                else:
                    self.debug("Reloading local file %s...", local_filename)
                    self.reload(idx, filename=local_filename, goto=line)
            elif last_answer in no_any:
                self.set_current_filename(local_filename)

    @ProtectTaurusMessageBox(title="An error occured trying to open macro(s)")
    def open_macro_libraries(self, macro_libraries):
        editorstack = self.editorStack()

        all_any = Qt.QMessageBox.YesToAll, Qt.QMessageBox.NoToAll
        yes_any = Qt.QMessageBox.Yes, Qt.QMessageBox.YesToAll
        no_any = Qt.QMessageBox.No, Qt.QMessageBox.NoToAll

        last_answer = Qt.QMessageBox.No
        for macro_library in macro_libraries:
            name, fname = macro_library.name, macro_library.file_path
            module = macro_library.module
            fname = fname[fname.index(osp.sep) + 1:]  # transform into relative path
            local_filename = osp.join(self._tmp_dir, fname)

            idx = editorstack.has_filename(local_filename)
            if idx is not None and last_answer not in all_any:
                last_answer = Qt.QMessageBox.question(self,
                    "Macro file '{0}' already opened".format(module),
                    "All changes to <b>{0}</b> will be lost."
                    "<br>Do you want to revert file from the server ignoring "
                    "any possible changes you (may) have made?".format(module),
                    Qt.QMessageBox.Yes | Qt.QMessageBox.YesToAll | \
                    Qt.QMessageBox.No | Qt.QMessageBox.NoToAll,
                    Qt.QMessageBox.No)

            if idx is None or last_answer in yes_any:
                if not self._prepare_path(macro_library.path):
                    Qt.QMessageBox.warning(self,
                        "Error trying to prepare '{0}'".format(module),
                        "An error occured trying to prepare '{0}'".format(module),
                        Qt.QMessageBox.Ok, Qt.QMessageBox.Ok)
                    continue
                _, code, line = self.get_macro_code(module)
                line = int(line)
                self.debug("Creating local file %s...", local_filename)
                fd = file(local_filename, "w")
                fd.write(code)
                fd.close()
                if idx is None:
                    self.debug("Loading local file %s...", local_filename)
                    self.load(local_filename, goto=line)
                else:
                    self.debug("Reloading local file %s...", local_filename)
                    self.reload(idx, filename=local_filename, goto=line)
            elif last_answer in no_any:
                self.set_current_filename(local_filename)

    @ProtectTaurusMessageBox(msg="A error occured trying to save")
    def on_save(self, apply=True):
        editorstack = self.editorStack()
        # Save the currently edited file
        if not editorstack.get_stack_count():
            return
        index = editorstack.get_stack_index()
        res = editorstack.save(index=index)
        file_info = editorstack.data[index]
        if not res:
            return
        local_filename = file_info.filename
        fd = file(local_filename, "r")
        code = fd.read()
        fd.close()
        remote_filename = local_filename[len(self._tmp_dir):]
        self.set_macro_code(remote_filename, code, apply)

    def on_revert(self):
        self.editorStack().revert()

    def reload_macro_lib(self, module_name):
        pass

    def set_macro_code(self, filename, code, apply=True):
        ms = self.getModelObj()
        if apply:
            apply = "true"
        else:
            apply = "false"
        return ms.SetMacroCode((filename, code, apply))

    def get_macro_code(self, module_name, macro_name=None):
        ms = self.getModelObj()
        pars = [ module_name ]
        if macro_name is not None:
            pars.append(macro_name)
        return ms.GetMacroCode(pars)

    def setModel(self, model_name):
        TaurusBaseWidget.setModel(self, model_name)
        self._elementTree.setModel(model_name)
        self._is_filesystem_prepared = False
        self.prepare_filesystem()

    def _prepare_path(self, p):
        if not self.prepare_filesystem():
            return False
        p = p[p.index(osp.sep) + 1:]  # transform into relative path
        p = osp.join(self._tmp_dir, p)
        if not osp.exists(p):
            os.makedirs(p)
        return True

    def prepare_filesystem(self):
        """Prepares a temporary directory to store the macro files locally"""
        if self._is_filesystem_prepared:
            return True
        ms = self.getModelObj()
        if ms is None:
            self.warning("Could not prepare local filesystem to store macros")
            return False
        ms_name = ms.getSimpleName().replace('/', '_')
        self._tmp_dir = tempfile.mkdtemp(prefix=ms_name, dir=self._base_tmp_dir)
        self._is_filesystem_prepared = True
        return True

    def closeEvent(self, event):
        if self._is_filesystem_prepared:
            shutil.rmtree(self._tmp_dir)
        TaurusBaseEditor.closeEvent(self, event)


def demo(model_name="MS_BL98"):
    test = SardanaEditor()
    test.resize(1000, 800)
    #test.load(__file__)
    test.setModel(model_name)
    test.show()
    return test

def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        parser.usage = "%prog [options] <macro server name>"
        app = Application(sys.argv, cmd_line_parser=parser,
                          app_name="Macro editor demo", app_version="1.0",
                          org_domain="Sardana", org_name="Tango community")

    args = app.get_command_line_args()

    if len(args) == 0:
        parser.error("must give a macroserver device name")
    else:
        w = demo(args[0])

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == "__main__":
    main()



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains a taurus text editor widget."""

__all__ = ["SardanaElementTreeWidget"]

__docformat__ = 'restructuredtext'

from taurus.core.taurusdevice import TaurusDevice
from taurus.qt.qtgui.tree import TaurusBaseTreeWidget

from sardana.taurus.qt.qtcore.tango.sardana.model import \
    SardanaBaseProxyModel, SardanaElementTypeModel


class SardanaElementTreeWidget(TaurusBaseTreeWidget):

    KnownPerspectives = { "Type" : {
                          "label" : "By type",
                          "icon" : ":/python-file.png",
                          "tooltip" : "View elements by type",
                          "model" : [SardanaBaseProxyModel, SardanaElementTypeModel],
                        },
    }
    DftPerspective = "Type"

    def getModelClass(self):
        return TaurusDevice

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseTreeWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.extra_sardana'
        ret['group'] = 'Taurus Sardana'
        ret['icon'] = ":/designer/listview.png"
        return ret


def main_SardanaTreeWidget(device):
    w = SardanaElementTreeWidget(with_navigation_bar=True)
    w.setWindowTitle("Sardana browser - " + device)
    w.setModel(device)
    w.setMinimumSize(400, 800)
    w.show()
    return w

def demo(device="V3"):
    """"""
    w = main_SardanaTreeWidget(device)
    return w

def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        app = Application(app_name="Pool element tree demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    args = app.get_command_line_args()
    if len(args) == 1:
        w = demo(device=args[0])
    else:
        w = demo()

    w.show()
    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == "__main__":
    main()



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Sardana libray. It defines a *slow* lock
class that provides additional debugging information"""

from __future__ import absolute_import

__all__ = ["SardanaLock"]

__docformat__ = 'restructuredtext'

import logging
import threading

_VERBOSE = True


def SardanaLock(verbose=None, name=None, lock=None):
    if verbose is None:
        verbose = _VERBOSE
    if verbose:
        return _SardanaLock(name=name, lock=lock)
    if lock is None:
        return threading.Lock()
    return lock


class _SardanaLock(object):
    """A sardana lock"""

    def __init__(self, name=None, lock=None, level=logging.DEBUG):
        name = name or self.__class__.__name__
        self.__name = name
        self.__logger = logging.getLogger(name=name)
        self.__level = level
        if lock is None:
            lock = threading.Lock()
        self.__block = lock
        self.__owner = None

    def __repr__(self):
        owner = self.__owner
        if owner is not None:
            owner = owner.name
        return "<%s owner=%r>" % (self.__name, owner)

    def owner_name(self):
        owner = self.__owner
        if owner is not None:
            return owner.name

    def _note(self, msg, *args):
        self.__logger.log(self.__level, msg, *args)

    def acquire(self, blocking=1):
        if __debug__:
            self._note("[START] acquire(%s) [owner=%s]", blocking,
                       self.owner_name())
        rc = self.__block.acquire(blocking)
        me = threading.current_thread()
        if rc:
            self.__owner = me
            state = "success"
        else:
            state = "failure"
        if __debug__:
            self._note("[ END ] acquire(%s) %s [owner=%s]", blocking, state,
                       self.owner_name())
        return rc

    __enter__ = acquire

    def release(self):
        if __debug__:
            self._note("[START] release() [owner=%s]", self.owner_name())
        self.__block.release()
        self.__owner = None
        if __debug__:
            self._note("[ END ] release() [owner=%s]", self.owner_name())

    def __exit__(self, t, v, tb):
        self.release()



##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""MCA related macros"""

__docformat__ = 'restructuredtext'

__all__ = ["mca_start", "mca_stop"]

from sardana.macroserver.macro import *

import array

class mca_start(Macro):
    """Starts an mca"""
    
    param_def = [
       ['mca', Type.ExpChannel, None, 'mca']
    ]
    
    def run(self, mca):
        name = mca.getName()
        mca.Start()
        o = "Starting " +  name
        self.output(o)

class mca_stop(Macro):
    """Stops an mca"""
    
    param_def = [
       ['mca', Type.ExpChannel, None, 'mca']
    ]
    
    def run(self, mca):
        name = mca.getName()
        mca.Abort()
        o = "Stopping " +  name
        self.output(o)



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Sardana library. It defines some
utility methods"""

from __future__ import absolute_import

__all__ = ["is_pure_str", "is_non_str_seq", "is_integer", "is_number",
           "is_bool", "check_type", "assert_type", "str_to_value",
           "is_callable", "translate_version_str2int",
           "translate_version_str2list"]

__docformat__ = 'restructuredtext'

import numpy
import numbers
import collections

from sardana.sardanadefs import DataType, DataFormat, DTYPE_MAP, R_DTYPE_MAP

__str_klasses = [str]
__int_klasses = [int, numpy.integer]
__number_klasses = [numbers.Number, numpy.number]

__DTYPE_MAP = dict(DTYPE_MAP)

__use_unicode = False
try:
    unicode
    __use_unicode = True
    __str_klasses.append(unicode)
    __DTYPE_MAP[unicode] = DataType.String
except:
    pass

__use_long = False
try:
    long
    __use_long = True
    __int_klasses.append(long)
    __DTYPE_MAP[long] = DataType.Integer
except:
    pass

__bool_klasses = [bool] + __int_klasses

__str_klasses = tuple(__str_klasses)
__int_klasses = tuple(__int_klasses)
__number_klasses = tuple(__number_klasses)
__bool_klasses = tuple(__bool_klasses)

def is_pure_str(obj):
    return isinstance(obj , __str_klasses)

def is_non_str_seq(obj):
    return isinstance(obj, collections.Sequence) and not is_pure_str(obj)

def is_integer(obj):
    return isinstance(obj, __int_klasses)

def is_number(obj):
    return isinstance(obj, __number_klasses)

def is_bool(obj):
    return isinstance(obj, __bool_klasses)

def is_callable(obj):
    return hasattr(obj, "__call__")

__METH_MAP = {
    DataType.Integer : is_integer,
    DataType.Double  : is_number,
    DataType.String  : is_pure_str,
    DataType.Boolean : is_bool,
}

def check_type(type_info, value):
    tinfo = __DTYPE_MAP.get(type_info, type_info)
    tmeth = __METH_MAP.get(tinfo, type_info)
    return tmeth(value)

def assert_type(type_info, value):
    ret = check_type(type_info, value)
    if not ret:
        expected = R_DTYPE_MAP[type_info]
        recv = type(value)
        try:
            expected = expected.__name__
        except:
            expected = str(expected)
        try:
            recv = recv.__name__
        except:
            recv = str(recv)
        raise TypeError("Expected %s, but received %s", expected, recv)
    return ret

_DTYPE_FUNC = {
    DataType.Integer : int,
    DataType.Double  : float,
    DataType.String  : str,
    DataType.Boolean : bool,
}

def str_to_value(value, dtype=DataType.Double, dformat=DataFormat.Scalar):
    f = _DTYPE_FUNC[dtype]
    if dformat == DataFormat.Scalar:
        ret = f(value)
    elif dformat == DataFormat.OneD:
        ret = [ f(v) for v in value ]
    elif dformat == DataFormat.TwoD:
        ret = []
        for v1 in value:
            ret.append([ f(v2) for v2 in v1 ])
    return ret

def translate_version_str2int(version_str):
    """Translates a version string in format x[.y[.z[...]]] into a 000000 number.
    Each part of version number can have up to 99 possibilities."""
    import math
    parts = version_str.split('.')
    i, v, l = 0, 0, len(parts)
    if not l: return v
    while i < 3:
        try:
            v += int(parts[i]) * int(math.pow(10, (2 - i) * 2))
            l -= 1
            i += 1
        except ValueError:
            return v
        if not l: return v
    return v

    try:
        v += 10000 * int(parts[0])
        l -= 1
    except ValueError:
        return v
    if not l: return v

    try:
        v += 100 * int(parts[1])
        l -= 1
    except ValueError:
        return v
    if not l: return v

    try:
        v += int(parts[0])
        l -= 1
    except ValueError:
        return v
    if not l: return v

def translate_version_str2list(version_str, depth=2):
    """Translates a version string in format 'x[.y[.z[...]]]' into a list of
    numbers"""
    if version_str is None:
        ver = depth * [0, ]
    else:
        ver = []
        for i in version_str.split(".")[:depth]:
            try:
                i = int(i)
            except:
                i = 0
            ver.append(i)
    return ver




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
historyviewer.py: 
"""
import copy

from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtcore.configuration import BaseConfigurableClass
from model import MacrosListModel

class HistoryMacrosViewer(TaurusWidget):
    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)
        self.setObjectName(self.__class__.__name__)
        self.registerConfigProperty("toXmlString", "fromXmlString", "history")
        self.initComponents()

    def initComponents(self):
        self.setLayout(Qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.list = HistoryMacrosList(self)
        self._model = MacrosListModel()
        self.list.setModel(self._model)

#####        self.registerConfigDelegate(self.list)
        self.layout().addWidget(self.list)

        actionBar = self.createActionBar()
        self.layout().addLayout(actionBar)

    def createActionBar(self):
        layout = Qt.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        deleteAllButton = Qt.QToolButton()
        deleteAllButton.setDefaultAction(self.list.removeAllAction)
        layout.addWidget(deleteAllButton)
        spacerItem = Qt.QSpacerItem(0, 0, Qt.QSizePolicy.Fixed, Qt.QSizePolicy.Expanding)
        layout.addItem(spacerItem)
        return layout

    def listElementUp(self):
        indexPos = self.list.currentIndex()
        if indexPos.isValid() and indexPos.row() >= 1:
            self.list.setCurrentIndex(indexPos.sibling(indexPos.row() - 1, indexPos.column()))
        else:
            self.selectFirstElement()

    def listElementDown(self):
        indexPos = self.list.currentIndex()
        if indexPos.isValid() and indexPos.row() < self._model.rowCount() - 1:
            self.list.setCurrentIndex(indexPos.sibling(indexPos.row() + 1, indexPos.column()))
        elif indexPos.row() == self._model.rowCount() - 1:
            return
        else:
            self.selectFirstElement()

    def addMacro(self, macroNode):
        self.list.insertMacro(macroNode)

    def toXmlString(self):
        return self.list.toXmlString()

    def fromXmlString(self, xmlString):
        self.list.fromXmlString(xmlString)
        historyList = self.list.model().list
        macroServerObj = self.getModelObj()
        if macroServerObj is None:
            return

        for macroNode in historyList:
            macroServerObj.fillMacroNodeAdditionalInfos(macroNode)

    def selectFirstElement(self):
        self.list.removeAllAction.setEnabled(True)
        self.list.setCurrentIndex(self._model.index(0))


    @classmethod
    def getQtDesignerPluginInfo(cls):
        return None


class HistoryMacrosList(Qt.QListView, BaseConfigurableClass):

    def __init__(self, parent):
        Qt.QListView.__init__(self, parent)
        self.setSelectionMode(Qt.QListView.SingleSelection)
        self.removeAllAction = Qt.QAction(Qt.QIcon(":/places/user-trash.svg"), "Remove all from history", self)
        self.connect(self.removeAllAction, Qt.SIGNAL("triggered()"), self.removeAllMacros)
        self.removeAllAction.setToolTip("Clicking this button will remove all macros from history.")
        self.removeAllAction.setEnabled(False)

    def currentChanged(self, current, previous):
        macro = copy.deepcopy(self.currentIndex().internalPointer())
        self.emit(Qt.SIGNAL("historySelected"), macro)
        Qt.QListView.currentChanged(self, current, previous)

    def mousePressEvent(self, e):
        clickedIndex = self.indexAt(e.pos())
        if clickedIndex.isValid():
            macro = copy.deepcopy(self.currentIndex().internalPointer())
            self.emit(Qt.SIGNAL("historySelected"), macro)
            self.removeAllAction.setEnabled(True)
        Qt.QListView.mousePressEvent(self, e)

    def focusInEvent(self, e):
        if self.model().rowCount() > 0:
            self.removeAllAction.setEnabled(True)
        else:
            self.removeAllAction.setEnabled(False)

    def insertMacro(self, macroNode):
        idx = self.model().insertRow(macroNode)
        self.setCurrentIndex(idx)
        self.removeAllAction.setEnabled(True)

    def removeAllMacros(self):
        self.selectAll()
        slist = sorted(self.selectedIndexes(), key=lambda index: index.row(), reverse=True)
        for index in slist:
            self.model().removeRow(index.row())
        self.removeAllAction.setEnabled(False)

    def toXmlString(self):
        return self.model().toXmlString()

    def fromXmlString(self, xmlString):
        self.model().fromXmlString(xmlString)


def test():
    import sys, taurus, time
    from  taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv)

    historyViewer = HistoryMacrosViewer()

    args = app.get_command_line_args()
    historyViewer.setModel(args[0])
    time.sleep(1)
    macroNode = historyViewer.getModelObj().getMacroNodeObj(str(args[1]))
    historyViewer.addMacro(macroNode)
    historyViewer.show()

    sys.exit(app.exec_())

if __name__ == "__main__":
    test()



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import json
import time
import threading

from lxml import etree

from PyTango import Util, DevFailed, Except, DevVoid, DevLong, \
    DevLong64, DevString, DevState, DevEncoded, \
    DevVarStringArray, ArgType, \
    READ, READ_WRITE, SCALAR, SPECTRUM

from taurus.core.util.log import DebugIt, LogFilter
from taurus.core.util.codecs import CodecFactory

from sardana import State, InvalidId, SardanaServer
from sardana.sardanaattribute import SardanaAttribute
from sardana.macroserver.macro import Macro
from sardana.macroserver.msdoor import BaseInputHandler
from sardana.macroserver.msexception import MacroServerException
from sardana.tango.core.util import throw_sardana_exception
from sardana.tango.core.attributehandler import AttributeLogHandler
from sardana.tango.core.SardanaDevice import SardanaDevice, SardanaDeviceClass
from sardana.macroserver.msexception import InputCancelled


class TangoInputHandler(BaseInputHandler):
    
    def __init__(self, door, attr):
        self._value = None
        self._input_waitting = False
        self._input_event = threading.Event()
        self._door = door
        self._attr = attr
    
    def input(self, input_data=None):
        if input_data is None:
            input_data = {}
        self.input_data = input_data
        timeout = input_data.get('timeout')
        input_data = json.dumps(input_data)
        self._value = None
        self._input_waitting = True
        try:
            self._door.set_attribute(self._attr, value=input_data)
            res = self.input_wait(timeout=timeout)
        finally:
            self._input_waitting = False
        
        if res is None or res.get('cancel', False):
            raise InputCancelled('Input cancelled by user')
        return res['input']

    def input_received(self, value):
        if not self._input_waitting:
            return
        self._value = json.loads(value)
        self._input_event.set()

    def input_wait(self, timeout=None):
        wait = self._input_event.wait(timeout)
        # if there was a timeout:
        # - set the value to the default value (if one exists)
        # - inform clients that timeout occured and they should not wait for 
        #   user input anymore
        if not self._input_event.is_set():
            if 'default_value' in self.input_data:
                self._value = dict(input=self.input_data['default_value'])
            self.send_input_timeout()
        self._input_event.clear()
        return self._value

    def send_input_timeout(self):
        idata = self.input_data
        input_data = dict(type="timeout", macro_id=idata['macro_id'])
        if 'default_value' in idata:
            input_data['default_value'] = idata['default_value']
        input_data = json.dumps(input_data)
        door = self._door
        door.set_attribute(self._attr, value=input_data)


class TangoFunctionHandler(object):

    def __init__(self, door, attr, module_name, format="bz2_pickle"):
        self.door = door
        self.attr = attr
        self.module_name = module_name
        self.format = format
        
    def handle(self, func_name, *args, **kwargs):
        codec = CodecFactory().getCodec(self.format)
        data = dict(type='function', func_name=func_name, args=args, kwargs=kwargs)
        event_value = codec.encode(('', data))
        self.door.set_attribute(self.attr, value=event_value)
    
    def __getattr__(self, name):
        def f(*args, **kwargs):
            full_name = self.module_name + "." + name
            return self.handle(full_name, *args, **kwargs)
        f.__name__ = name
        return f


class TangoPylabHandler(TangoFunctionHandler):

    def __init__(self, door, attr, format="bz2_pickle"):
        TangoFunctionHandler.__init__(self, door, attr, "pylab",
                                      format="bz2_pickle")

         
class TangoPyplotHandler(TangoFunctionHandler):

    def __init__(self, door, attr, format="bz2_pickle"):
        TangoFunctionHandler.__init__(self, door, attr, "pyplot",
                                      format="bz2_pickle")
        

class Door(SardanaDevice):

    def __init__(self, dclass, name):
        SardanaDevice.__init__(self, dclass, name)
        self._last_result = ()
        self._input_handler = None

    def init(self, name):
        SardanaDevice.init(self, name)
        self._door = None
        self._macro_server_device = None

    def get_door(self):
        return self._door

    def set_door(self, door):
        self._door = door

    door = property(get_door, set_door)

    @property
    def macro_server_device(self):
        return self._macro_server_device

    @property
    def macro_server(self):
        return self.door.macro_server

    def delete_device(self):
        if self.getRunningMacro():
            self.debug("aborting running macro")
            self.macro_executor.abort()
        
        for handler, filter, format in self._handler_dict.values():
            handler.finish()

        door = self.door
        if door is not None:
            door.remove_listener(self.on_door_changed)

    @DebugIt()
    def init_device(self):
        SardanaDevice.init_device(self)
        levels = 'Critical', 'Error', 'Warning', 'Info', 'Output', 'Debug'
        detect_evts = ()
        non_detect_evts = ['State', 'Status', 'Result', 'RecordData',
                           'MacroStatus', 'Input'] + list(levels)
        self.set_change_events(detect_evts, non_detect_evts)

        util = Util.instance()
        db = util.get_database()
        
        # Find the macro server for this door
        macro_servers = util.get_device_list_by_class("MacroServer")
        if self.MacroServerName is None:
            self._macro_server_device = macro_servers[0]
        else:
            ms_name = self.MacroServerName.lower()
            for ms in macro_servers:
                if ms.get_name().lower() == ms_name or \
                   ms.alias.lower() == ms_name:
                    self._macro_server_device = ms
                    break
        
        # support for old doors which didn't have ID
        if self.Id == InvalidId:
            self.Id = self.macro_server_device.macro_server.get_new_id()
            db.put_device_property(self.get_name(), dict(Id=self.Id))
        
        door = self.door
        if door is None:
            full_name = self.get_name()
            name = full_name
            macro_server = self.macro_server_device.macro_server
            self.door = door = \
                macro_server.create_element(type="Door", name=name,
                                            full_name=full_name, id=self.Id)
            self._setupLogHandlers(levels)

        multi_attr = self.get_device_attr()

        input_attr = multi_attr.get_attr_by_name('Input')
        self._input_handler = ih = TangoInputHandler(self, input_attr)
        door.set_input_handler(ih)
        
        recorddata_attr = multi_attr.get_attr_by_name('RecordData')
        self._pylab_handler = pylabh = TangoPylabHandler(self, recorddata_attr)
        door.set_pylab_handler(pylabh)
        
        self._pyplot_handler = pyploth = TangoPyplotHandler(self, recorddata_attr)
        door.set_pyplot_handler(pyploth)

        door.add_listener(self.on_door_changed)
        self.set_state(DevState.ON)

    def _setupLogHandlers(self, levels):
        self._handler_dict = {}
        for level in levels:
            handler = AttributeLogHandler(self, level,
                                          max_buff_size=self.MaxMsgBufferSize)
            filter = LogFilter(level=getattr(self, level))
            handler.addFilter(filter)
            self.addLogHandler(handler)
            format = None
            self._handler_dict[level] = handler, filter, format
    
    def on_door_changed(self, event_source, event_type, event_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return
        
        timestamp = time.time()
        
        name = event_type.name.lower()
        
        multi_attr = self.get_device_attr()
        try:
            attr = multi_attr.get_attr_by_name(name)
        except DevFailed:
            return
        
        if name == "state":
            event_value = self.calculate_tango_state(event_value)
        elif name == "status":
            event_value = self.calculate_tango_status(event_value)
        elif name == "recorddata":
            format, value = event_value
            codec = CodecFactory().getCodec(format)
            event_value = codec.encode(('', value))
        else:
            if isinstance(event_value, SardanaAttribute):
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                timestamp = event_value.timestamp
                event_value = event_value.value
            
            if attr.get_data_type() == ArgType.DevEncoded:
                codec = CodecFactory().getCodec('json')
                event_value = codec.encode(('', event_value))
        self.set_attribute(attr, value=event_value, timestamp=timestamp)
    
    @property
    def macro_executor(self):
        return self.door.macro_executor
    
    def getRunningMacro(self):
        return self.door.running_macro
    
    def always_executed_hook(self):
        pass
    
    def read_attr_hardware(self,data):
        pass
    
    def readLogAttr(self, attr):
        name = attr.get_name()
        handler, filter, format = self._handler_dict[name]
        handler.read(attr)
    
    read_Critical = read_Error = read_Warning = read_Info = read_Output = \
        read_Debug = read_Trace = readLogAttr 
    
    def read_Input(self, attr):
        attr.set_value('')
    
    def write_Input(self, attr):
        value = attr.get_write_value()
        self.door.get_input_handler().input_received(value)
    
    #@DebugIt()
    def read_ElementList(self, attr):
        element_list = self.macro_server_device.getElementList()
        attr.set_value(*element_list)
    
    def sendRecordData(self, format, data):
        self.push_change_event('RecordData', format, data)

    def getLogAttr(self, name):
        return self._handler_dict.get(name)
    
    def read_Result(self, attr):
        #    Add your own code here
        attr.set_value(self._last_result)
    
    def read_RecordData(self, attr):
        try:
            macro_data = self.door.get_macro_data()
            codec = CodecFactory().getCodec('bz2_pickle')
            data = codec.encode(('', macro_data))
        except MacroServerException, mse:
            throw_sardana_exception(mse)

        attr.set_value(*data)
    
    def read_MacroStatus(self, attr):
        attr.set_value('', '')
    
    def Abort(self):
        self.debug("Abort is deprecated. Use StopMacro instead")
        return self.StopMacro()
    
    def AbortMacro(self):
        self.debug("Aborting")
        self.macro_executor.abort()
        self.debug("Finished aborting")
        
    def is_Abort_allowed(self):
        return True

    def PauseMacro(self):
        macro = self.getRunningMacro()
        if macro is None:
            print "Unable to pause Null macro"
            return
        self.macro_executor.pause()

    def is_PauseMacro_allowed(self):
        return self.get_state() == Macro.Running

    def StopMacro(self):
        macro = self.getRunningMacro()
        if macro is None:
            return
        self.debug("stopping macro %s" % macro._getDescription())
        self.macro_executor.stop()
    
    def is_StopMacro_allowed(self):
        return self.get_state() == Macro.Running
    
    def ResumeMacro(self):
        macro = self.getRunningMacro()
        if macro is None:
            return
        self.debug("resume macro %s" % macro._getDescription())
        self.macro_executor.resume()
        
    def is_ResumeMacro_allowed(self):
        return self.get_state() == Macro.Pause
    
    def RunMacro(self, par_str_list):
        #first empty all the buffers
        for handler, filter, fmt in self._handler_dict.values():
            handler.clearBuffer()
        
        if len(par_str_list) == 0:
            return []
        
        xml_seq = self.door.run_macro(par_str_list, asynch=True)
        return [etree.tostring(xml_seq, pretty_print=False)]

    def is_RunMacro_allowed(self):
        return self.get_state() in [Macro.Finished, Macro.Abort]

    def SimulateMacro(self, par_str_list):
        raise Exception("Not implemented yet")
    
    def GetMacroEnv(self, argin):
        macro_name = argin[0]
        if len(argin) > 1:
            macro_env = argin[1:]
        else:
            macro_env = self.door.get_macro_class_info(macro_name).env
        env = self.door.get_env(macro_env, macro_name=macro_name)
        ret = []
        for k,v in env.iteritems():
            ret.extend((k,v))
        return ret
    
    def is_GetMacroEnv_allowed(self):
        return self.get_state() in [Macro.Finished, Macro.Abort]


class DoorClass(SardanaDeviceClass):

    #    Class Properties
    class_property_list = {
        }

    #    Device Properties
    device_property_list = {
        'Id': [DevLong64, "Internal ID", [ InvalidId ] ],
        'MaxMsgBufferSize':
            [DevLong,
            'Maximum size for the Output, Result, Error, Warning, Debug and '
            'Info buffers',
            [512] ],
        'MacroServerName':
            [DevString,
            'Name of the macro server device to connect to. [default: None, '
            'meaning connect to the first registered macroserver',
            None ],
        }

    #    Command definitions
    cmd_list = {
        'Abort':
            [ [ DevVoid, ""],
              [ DevVoid, ""] ],
        'PauseMacro':
            [ [DevVoid, ""],
              [DevVoid, ""] ],
        'AbortMacro':
            [ [DevVoid, ""],
              [DevVoid, ""] ],
        'StopMacro':
            [ [DevVoid, ""],
              [DevVoid, ""] ],
        'ResumeMacro':
            [ [DevVoid, ""],
              [DevVoid, ""] ],
        'RunMacro':
            [ [DevVarStringArray, 'Macro name and parameters'],
              [DevVarStringArray, 'Macro Result']],
        'SimulateMacro':
            [ [DevVarStringArray, 'Macro name and parameters'],
              [DevVarStringArray, 'Macro statistics']],
        'GetMacroEnv':
            [ [ DevVarStringArray, 'Macro name followed by an ' \
                'optional list of environment names' ],
              [ DevVarStringArray, 'Macro environment as a list of '\
                'pairs keys, value'] ],
#        'ReloadMacro':
#            [[DevVarStringArray, "Macro(s) name(s)"],
#            [DevVarStringArray, "[OK] if successfull or a traceback " \
#                "if there was a error (one string with complete traceback of " \
#                "each error)"]],
#        'ReloadMacroLib':
#            [[DevVarStringArray, "MacroLib(s) name(s)"],
#            [DevVarStringArray, "[OK] if successfull or a traceback " \
#                "if there was a error (one string with complete traceback of " \
#                "each error)"]],
        }

    #    Attribute definitions
    attr_list = {
        'Result'        : [ [ DevString, SPECTRUM, READ, 512],
                            { 'label'     : 'Result for the last macro', } ],
        'Critical'      : [ [ DevString, SPECTRUM, READ, 512],
                            { 'label'     : 'Macro critical error message', } ],
        'Error'         : [ [ DevString, SPECTRUM, READ, 512],
                            { 'label'     : 'Macro error message', } ],
        'Warning'       : [ [ DevString, SPECTRUM, READ, 512],
                            { 'label'     : 'Macro warning message', } ],
        'Info'          : [ [ DevString, SPECTRUM, READ, 512],
                            { 'label'     : 'Macro information message', } ],
        'Debug'         : [ [ DevString, SPECTRUM, READ, 512],
                            { 'label'     : 'Macro debug message', } ],
        'Output'        : [ [ DevString, SPECTRUM, READ, 512],
                            { 'label'     : 'Macro output message', } ],
        'Input'         : [ [ DevString, SCALAR, READ_WRITE],
                            { 'label'     : 'Macro input prompt', } ],
        'RecordData'    : [ [ DevEncoded, SCALAR, READ],
                            { 'label'     : 'Record Data', } ],
        'MacroStatus'   : [ [ DevEncoded, SCALAR, READ],
                            { 'label'     : 'Macro Status', } ],
        'ElementList'   : [ [ DevEncoded, SCALAR, READ],
                            { 'label':"Element list",
                              'description' : 'the list of all elements (a '
                                              'JSON encoded dict)', } ],
        }



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" """

__all__ = ["ZeroDExpChannel", "ZeroDExpChannelClass"]

__docformat__ = 'restructuredtext'

import time

from PyTango import Except
from PyTango import DevVoid, DevDouble, DevString
from PyTango import DispLevel, DevState, AttrQuality
from PyTango import READ, READ_WRITE, SCALAR, SPECTRUM

from taurus.core.util.log import DebugIt

from sardana import State, DataFormat, SardanaServer
from sardana.sardanaattribute import SardanaAttribute
from sardana.pool.controller import ZeroDController, Type
from sardana.tango.core.util import to_tango_type_format

from sardana.tango.pool.PoolDevice import PoolElementDevice, \
    PoolElementDeviceClass


class ZeroDExpChannel(PoolElementDevice):

    def __init__(self, dclass, name):
        PoolElementDevice.__init__(self, dclass, name)

    def init(self, name):
        PoolElementDevice.init(self, name)

    def get_zerod(self):
        return self.element

    def set_zerod(self, zerod):
        self.element = zerod

    zerod = property(get_zerod, set_zerod)

    @DebugIt()
    def delete_device(self):
        PoolElementDevice.delete_device(self)
        zerod = self.zerod
        if zerod is not None:
            zerod.remove_listener(self.on_zerod_changed)

    @DebugIt()
    def init_device(self):
        PoolElementDevice.init_device(self)
        zerod = self.zerod
        if zerod is None:
            full_name = self.get_full_name()
            name = self.alias or full_name
            self.zerod = zerod = \
                self.pool.create_element(type="ZeroDExpChannel", name=name,
                    full_name=full_name, id=self.Id, axis=self.Axis,
                    ctrl_id=self.Ctrl_id)
        zerod.add_listener(self.on_zerod_changed)

        ## force a state read to initialize the state attribute
        #state = zerod.state
        self.set_state(DevState.ON)

    def on_zerod_changed(self, event_source, event_type, event_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return

        timestamp = time.time()
        name = event_type.name
        quality = AttrQuality.ATTR_VALID
        priority = event_type.priority
        error = None
        attr = self.get_device_attr().get_attr_by_name(name)

        if name == "state":
            event_value = self.calculate_tango_state(event_value)
        elif name == "status":
            event_value = self.calculate_tango_status(event_value)
        else:
            if isinstance(event_value, SardanaAttribute):
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                timestamp = event_value.timestamp
                event_value = event_value.value

            if name == "value":
                state = self.zerod.get_state()
                if state == State.Moving:
                    quality = AttrQuality.ATTR_CHANGING
        self.set_attribute(attr, value=event_value, timestamp=timestamp,
                           quality=quality, priority=priority, error=error,
                           synch=False)

    def always_executed_hook(self):
        #state = to_tango_state(self.zerod.get_state(cache=False))
        pass

    def read_attr_hardware(self, data):
        pass

    def get_dynamic_attributes(self):
        cache_built = hasattr(self, "_dynamic_attributes_cache")

        std_attrs, dyn_attrs = \
            PoolElementDevice.get_dynamic_attributes(self)

        if not cache_built:
            # For value attribute, listen to what the controller says for data
            # type (between long and float)
            value = std_attrs.get('value')
            if value is not None:
                attr_name, data_info, attr_info = value
                ttype, _ = to_tango_type_format(attr_info.dtype)
                data_info[0][0] = ttype

                # Add manually a 'CurrentValue' with the same time as 'Value'
                attr_name = 'CurrentValue'
                attr_info = attr_info.copy()
                attr_info.description = attr_name
                std_attrs[attr_name] = [attr_name, data_info, attr_info]

        return std_attrs, dyn_attrs

    def initialize_dynamic_attributes(self):
        attrs = PoolElementDevice.initialize_dynamic_attributes(self)

        detect_evts = "value",
        non_detect_evts = ()

        for attr_name in detect_evts:
            if attr_name in attrs:
                self.set_change_event(attr_name, True, True)
        for attr_name in non_detect_evts:
            if attr_name in attrs:
                self.set_change_event(attr_name, True, False)

    def read_Value(self, attr):
        zerod = self.zerod
        value = zerod.get_accumulated_value()
        quality = None
        if self.get_state() == State.Moving:
            quality = AttrQuality.ATTR_CHANGING
        self.set_attribute(attr, value=value.value, quality=quality, priority=0)

    def read_CurrentValue(self, attr):
        zerod = self.zerod
        #use_cache = ct.is_action_running() and not self.Force_HW_Read
        use_cache = self.get_state() == State.Moving and not self.Force_HW_Read
        value = zerod.get_current_value(cache=use_cache, propagate=0)
        if value.error:
            Except.throw_python_exception(*value.exc_info)
        quality = None
        state = zerod.get_state(cache=use_cache, propagate=0)
        if state == State.Moving:
            quality = AttrQuality.ATTR_CHANGING
        self.set_attribute(attr, value=value.value, quality=quality,
                           priority=0, timestamp=value.timestamp)

    def Start(self):
        self.zerod.start_acquisition()

    def read_ValueBuffer(self, attr):
        attr.set_value(self.zerod.get_value_buffer())

    def read_TimeBuffer(self, attr):
        attr.set_value(self.zerod.get_time_buffer())

    def read_CumulationType(self, attr):
        attr.set_value(self.zerod.get_cumulation_type())

    def write_CumulationType(self, attr):
        self.zerod.set_cumulation_type(attr.get_write_value())

    def _is_allowed(self, req_type):
        return PoolElementDevice._is_allowed(self, req_type)

    is_Value_allowed = _is_allowed
    is_CurrentValue_allowed = _is_allowed
    is_CumulationType_allowed = _is_allowed
    is_ValueBuffer_allowed = _is_allowed
    is_TimeBuffer_allowed = _is_allowed


_DFT_VALUE_INFO = ZeroDController.standard_axis_attributes['Value']
_DFT_VALUE_TYPE, _DFT_VALUE_FORMAT = to_tango_type_format(_DFT_VALUE_INFO[Type], DataFormat.Scalar)


class ZeroDExpChannelClass(PoolElementDeviceClass):

    #    Class Properties
    class_property_list = {
    }

    #    Device Properties
    device_property_list = {
    }
    device_property_list.update(PoolElementDeviceClass.device_property_list)

    #    Command definitions
    cmd_list = {
        'Start' :   [ [DevVoid, ""], [DevVoid, ""] ],
    }
    cmd_list.update(PoolElementDeviceClass.cmd_list)

    #    Attribute definitions
    attr_list = {
        'ValueBuffer'    : [ [ DevDouble, SPECTRUM, READ, 16 * 1024 ] ],
        'TimeBuffer'     : [ [ DevDouble, SPECTRUM, READ, 16 * 1024 ] ],
        'CumulationType' : [ [ DevString, SCALAR, READ_WRITE ],
                             { 'Memorized'     : "true",
                               'label'         : "Cumulation Type",
                               'Display level' : DispLevel.EXPERT } ],
    }
    attr_list.update(PoolElementDeviceClass.attr_list)

    standard_attr_list = {
        'Value'          : [ [ _DFT_VALUE_TYPE, SCALAR, READ, ],
                             { 'abs_change' : '1.0', } ],
    }
    standard_attr_list.update(PoolElementDeviceClass.standard_attr_list)

    def _get_class_properties(self):
        ret = PoolElementDeviceClass._get_class_properties(self)
        ret['Description'] = "0D experimental channel device class"
        ret['InheritedFrom'].insert(0, 'PoolElementDevice')
        return ret



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""""""

__all__ = ["wraps", "wrapped", "is_wrapping", "is_wrapped"]

import weakref
from functools import wraps as _wraps

__WRAPPED = "__wrapped__"
__WRAPPER = "__wrapper__"

def wraps(wrapped, *args, **kwargs):
    """A wrap decorator which stores in the returned function a reference to
    the wrapped function (in member '__wrapped__')"""
    wrapper = _wraps(wrapped, *args, **kwargs)
    setattr(wrapper, __WRAPPED, weakref.ref(wrapped))
    setattr(wrapped, __WRAPPER, weakref.ref(wrapper))
    return wrapper

def is_wrapping(wrapper):
    """Determines if the given callable is a wrapper for another callable"""
    return hasattr(wrapper, __WRAPPED)

def is_wrapped(wrapped):
    """Determines if the given callable is being wrapped by another callable"""
    return hasattr(wrapped, __WRAPPER)

def wrapped(wrapper, recursive=True):
    """Returns the wrapped function around the given wrapper. If the given
    callable is not "wrapping" any function, the wrapper itself is returned"""
    if is_wrapping(wrapper):
        _wrapped = wrapper.__wrapped__()
    else:
        return wrapper

    if recursive:
        return wrapped(_wrapped)
    return _wrapped




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for instrument"""

__all__ = ["PoolInstrument"]

__docformat__ = 'restructuredtext'

import weakref

from sardana import ElementType

from sardana.pool.poolobject import PoolObject


class PoolInstrument(PoolObject):

    def __init__(self, **kwargs):
        self._parent_instrument = kwargs.pop("parent")
        self._instrument_class = kwargs.pop("klass")
        if self._parent_instrument:
            self._parent_instrument = weakref.ref(self._parent_instrument)
        self._child_instruments = {}
        self._elements = {}
        kwargs['elem_type'] = ElementType.Instrument
        super(PoolInstrument, self).__init__(**kwargs)

    def get_parent(self):
        return self.get_parent_instrument()

    def serialize(self, *args, **kwargs):
        kwargs = PoolObject.serialize(self, *args, **kwargs)
        kwargs['klass'] = self.instrument_class
        if self.parent_instrument is not None:
            kwargs['parent_instrument'] = self.parent_instrument.full_name
        else:
            kwargs['parent_instrument'] = None
        return kwargs

    def get_instrument_class(self):
        return self._instrument_class

    def add_instrument(self, instrument):
        self._child_instruments[instrument.id] = instrument

    def remove_instrument(self, instrument):
        del self._child_instruments[instrument.id]

    def get_instruments(self):
        return self._child_instruments.values()

    def set_parent_instrument(self, instrument):
        if instrument:
            self._parent_instrument = weakref.ref(instrument)
        else:
            self._parent_instrument = None

    def get_parent_instrument(self):
        if self.has_parent_instrument():
            return self._parent_instrument()

    def has_parent_instrument(self):
        return self._parent_instrument is not None

    def add_element(self, element):
        self._elements[element.id] = weakref.ref(element)

    def remove_element(self, element):
        del self._elements[element.id]

    def get_elements(self):
        return [ e() for e in self._elements.values() ]

    def has_instruments(self):
        return len(self._child_instruments) > 0

    def has_elements(self):
        return len(self._elements) > 0

    instruments = property(get_instruments)
    elements = property(get_elements)
    instrument_class = property(get_instrument_class)
    parent_instrument = property(get_parent_instrument)



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the definition of the macroserver parameters for
macros"""

__all__ = ["WrongParam", "MissingParam", "UnknownParamObj", "WrongParamType",
           "TypeNames", "Type", "ParamType", "ParamRepeat", "ElementParamType",
           "ElementParamInterface", "AttrParamType", "AbstractParamTypes",
           "ParamDecoder"]

__docformat__ = 'restructuredtext'

from taurus.core.util.containers import CaselessDict

from sardana import ElementType, INTERFACES_EXPANDED
from sardana.macroserver.msbase import MSBaseObject
from sardana.macroserver.msexception import MacroServerException, \
    UnknownMacro, UnknownMacroLibrary


class WrongParam(MacroServerException):

    def __init__(self, *args):
        MacroServerException.__init__(self, *args)
        self.type = 'Wrong parameter'


class MissingParam(WrongParam):

    def __init__(self, *args):
        WrongParam.__init__(self, *args)
        self.type = 'Missing parameter'


class UnknownParamObj(WrongParam):

    def __init__(self, *args):
        WrongParam.__init__(self, *args)
        self.type = 'Unknown parameter'


class WrongParamType(WrongParam):

    def __init__(self, *args):
        WrongParam.__init__(self, *args)
        self.type = 'Unknown parameter type'


class TypeNames:
    """Class that holds the list of registered macro parameter types"""

    def __init__(self):
        self._type_names = {}
        self._pending_type_names = {}

    def addType(self, name):
        """Register a new macro parameter type"""
        setattr(self, name, name)
        self._type_names[name] = name
        if name in self._pending_type_names:
            del self._pending_type_names[name]

    def removeType(self, name):
        """remove a macro parameter type"""
        delattr(self, name)
        try:
            del self._type_names[name]
        except ValueError:
            pass

    def __str__(self):
        return str(self._type_names.keys())

#    def __getattr__(self, name):
#        if name not in self._pending_type_names:
#            self._pending_type_names[name] = name
#        return self._pending_type_names[name]


# This instance of TypeNames is intended to provide access to types to the
# Macros in a "Type.Motor" fashion
Type = TypeNames()


class ParamType(MSBaseObject):

    All = 'All'

    # Capabilities
    ItemList = 'ItemList'
    ItemListEvents = 'ItemListEvents'

    capabilities = []

    type_class = str

    def __init__(self, macro_server, name):
        MSBaseObject.__init__(self, name=name, full_name=name,
                              macro_server=macro_server,
                              elem_type=ElementType.ParameterType)

    def getName(self):
        return self.name

    def getObj(self, str_repr):
        return self.type_class(str_repr)

    @classmethod
    def hasCapability(cls, cap):
        return cap in cls.capabilities

    def serialize(self, *args, **kwargs):
        kwargs = MSBaseObject.serialize(self, *args, **kwargs)
        kwargs['composed'] = False
        return kwargs


class ParamRepeat(object):
    # opts: min, max
    def __init__(self, *param_def, **opts):
        self.param_def = param_def
        self.opts = {'min': 1, 'max': None}
        self.opts.update(opts)
        self._obj = list(param_def)
        self._obj.append(self.opts)

    def items(self):
        return self.opts.items()

    def __getattr__(self, name):
        return self.opts[name]

    def obj(self):
        return self._obj


class ElementParamType(ParamType):

    capabilities = ParamType.ItemList, ParamType.ItemListEvents

    def __init__(self, macro_server, name):
        ParamType.__init__(self, macro_server, name)

    def accepts(self, elem):
        return elem.getType() == self._name

    def getObj(self, name, pool=ParamType.All, cache=False):
        macro_server = self.macro_server
        if pool == ParamType.All:
            pools = macro_server.get_pools()
        else:
            pools = macro_server.get_pool(pool),
        for pool in pools:
            elem_info = pool.getObj(name, elem_type=self._name)
            if elem_info is not None and self.accepts(elem_info):
                return elem_info
        # not a pool object, maybe it is a macro server object (perhaps a macro
        # code or a macro library
        try:
            return macro_server.get_macro(name)
        except UnknownMacro:
            pass

        try:
            return macro_server.get_macro_lib(name)
        except UnknownMacroLibrary:
            pass
        # neither pool nor macroserver contains any element with this name
        raise UnknownParamObj('%s with name %s does not exist' % \
                              (self._name, name))

    def getObjDict(self, pool=ParamType.All, cache=False):
        macro_server = self.macro_server
        objs = CaselessDict()
        if pool == ParamType.All:
            pools = macro_server.get_pools()
        else:
            pools = macro_server.get_pool(pool),
        for pool in pools:
            for elem_info in pool.getElements():
                if self.accepts(elem_info):
                    objs[elem_info.name] = elem_info
        for macro_lib_name, macro_lib in macro_server.get_macros().items():
            if self.accepts(macro_lib):
                objs[macro_lib_name] = macro_lib
        for macro_name, macro in macro_server.get_macros().items():
            if self.accepts(macro):
                objs[macro_name] = macro

        return objs

    def getObjListStr(self, pool=ParamType.All, cache=False):
        obj_dict = self.getObjDict(pool=pool, cache=cache)
        return obj_dict.keys()

    def getObjList(self, pool=ParamType.All, cache=False):
        obj_dict = self.getObjDict(pool=pool, cache=cache)
        return obj_dict.values()

    def serialize(self, *args, **kwargs):
        kwargs = ParamType.serialize(self, *args, **kwargs)
        kwargs['composed'] = True
        return kwargs


class ElementParamInterface(ElementParamType):

    def __init__(self, macro_server, name):
        ElementParamType.__init__(self, macro_server, name)
        bases, doc = INTERFACES_EXPANDED.get(name)
        self._interfaces = bases

    def accepts(self, elem):
        elem_type = elem.getType()
        elem_interfaces = INTERFACES_EXPANDED.get(elem_type)[0]
        if elem_interfaces is None:
            return ElementParamType.accepts(self, elem)
        return self._name in elem_interfaces

    def getObj(self, name, pool=ParamType.All, cache=False):
        macro_server = self.macro_server
        if pool == ParamType.All:
            pools = macro_server.get_pools()
        else:
            pools = macro_server.get_pool(pool),
        for pool in pools:
            elem_info = pool.getElementWithInterface(name, self._name)
            if elem_info is not None and self.accepts(elem_info):
                return elem_info
        # not a pool object, maybe it is a macro server object (perhaps a macro
        # class or a macro library
        try:
            return macro_server.get_macro(name)
        except UnknownMacro:
            pass

        try:
            return macro_server.get_macro_lib(name)
        except UnknownMacroLibrary:
            pass
        # neither pool nor macroserver contains any element with this name
        raise UnknownParamObj('%s with name %s does not exist' % \
                              (self._name, name))

    def getObjDict(self, pool=ParamType.All, cache=False):
        macro_server = self.macro_server
        objs = CaselessDict()
        if macro_server.is_macroserver_interface(self._name):
            return macro_server.get_elements_with_interface(self._name)

        if pool == ParamType.All:
            pools = macro_server.get_pools()
        else:
            pools = macro_server.get_pool(pool),
        for pool in pools:
            for elem_info in pool.getElementsWithInterface(self._name).values():
                if self.accepts(elem_info):
                    objs[elem_info.name] = elem_info
        return objs

    def getObjListStr(self, pool=ParamType.All, cache=False):
        obj_dict = self.getObjDict(pool=pool, cache=cache)
        return obj_dict.keys()

    def getObjList(self, pool=ParamType.All, cache=False):
        obj_dict = self.getObjDict(pool=pool, cache=cache)
        return obj_dict.values()


class AttrParamType(ParamType):
    pass


AbstractParamTypes = ParamType, ElementParamType, ElementParamInterface, AttrParamType


class ParamDecoder:

    def __init__(self, door, macro_meta, param_str_list):
        self.door = door
        self.macro_meta = macro_meta
        self.param_str_list = param_str_list
        self.param_list = None
        self.decode()

    @property
    def type_manager(self):
        return self.door.type_manager

    def decode(self):
        macro_meta = self.macro_meta
        macro_type = macro_meta.get_type()
        pars_str = self.param_str_list[1:]
        pars_def = macro_meta.get_parameter_definition()
        if macro_type == ElementType.MacroClass:
            _, self.param_list = self.decodeNormal(pars_str, pars_def)
        elif macro_type == ElementType.MacroFunction:
            if macro_meta.function.param_def is None:
                self.param_list = self.param_str_list[1:]
            else:
                _, self.param_list = self.decodeNormal(pars_str, pars_def)
        return self.param_list

    def decodeNormal(self, str_list, def_list):
        str_len = len(str_list)
        par_len = len(def_list)
        obj_list = []
        str_idx = 0
        for i, par_def in enumerate(def_list):
            name, type_class, def_val, desc = par_def
            if str_idx == str_len:
                if def_val is None:
                    if not isinstance(type_class, ParamRepeat):
                        raise MissingParam, "'%s' not specified" % name
                    elif isinstance(type_class, ParamRepeat):
                        min = par_def[1].opts['min']
                        if min > 0:
                            raise WrongParam, "'%s' demands at least %d values" % (name, min)
                new_obj_list = []
                if not def_val is None:
                    new_obj_list.append(def_val)
            else:
                if isinstance(type_class, ParamRepeat):
                    data = self.decodeRepeat(str_list[str_idx:], par_def)
                    dec_token, new_obj_list = data
                else:
                    type_manager = self.type_manager
                    type_name = type_class
                    type_class = type_manager.getTypeClass(type_name)
                    par_type = type_manager.getTypeObj(type_name)
                    par_str = str_list[str_idx]
                    try:
                        val = par_type.getObj(par_str)
                    except ValueError, e:
                        raise WrongParamType, e.message
                    except UnknownParamObj, e:
                        raise WrongParam, e.message
                    if val is None:
                        msg = 'Could not create %s parameter "%s" for "%s"' % \
                              (par_type.getName(), name, par_str)
                        raise WrongParam, msg
                    dec_token = 1
                    new_obj_list = [val]
                str_idx += dec_token
            obj_list += new_obj_list
        return str_idx, obj_list

    def decodeRepeat(self, str_list, par_def):
        name, rep_data, def_val, desc = par_def
        min_rep = rep_data.min
        max_rep = rep_data.max
        dec_token = 0
        obj_list = []
        rep_nr = 0
        while dec_token < len(str_list):
            if max_rep is not None and rep_nr == max_rep:
                break
            new_token, new_obj_list = self.decodeNormal(str_list[dec_token:],
                                                        rep_data.param_def)
            dec_token += new_token
            if len(new_obj_list) == 1:
                new_obj_list = new_obj_list[0]
            obj_list.append(new_obj_list)
            rep_nr += 1
        if rep_nr < min_rep:
            msg = 'Found %d repetitions of param %s, min is %d' % \
                  (rep_nr, name, min_rep)
            raise RuntimeError, msg
        return dec_token, obj_list

    def getParamList(self):
        return self.param_list

    def __getattr__(self, name):
        return getattr(self.param_list, name)



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""The sardana tango measurement group module"""

__all__ = ["MeasurementGroup", "MeasurementGroupClass"]

__docformat__ = 'restructuredtext'

import sys
import time

from PyTango import Except, DevVoid, DevLong, DevDouble, DevString, \
    DispLevel, DevState, AttrQuality, \
    READ_WRITE, SCALAR

from taurus.core.util.codecs import CodecFactory
from taurus.core.util.log import DebugIt

from sardana import State, SardanaServer
from sardana.sardanaattribute import SardanaAttribute
from sardana.pool import AcqMode
from sardana.tango.core.util import exception_str
from sardana.tango.pool.PoolDevice import PoolGroupDevice, PoolGroupDeviceClass


class MeasurementGroup(PoolGroupDevice):

    def __init__(self, dclass, name):
        PoolGroupDevice.__init__(self, dclass, name)

    def init(self, name):
        PoolGroupDevice.init(self, name)

    def get_measurement_group(self):
        return self.element

    def set_measurement_group(self, measurement_group):
        self.element = measurement_group

    measurement_group = property(get_measurement_group, set_measurement_group)

    @DebugIt()
    def delete_device(self):
        PoolGroupDevice.delete_device(self)
        mg = self.measurement_group
        if mg is not None:
            mg.remove_listener(self.on_measurement_group_changed)

    @DebugIt()
    def init_device(self):
        PoolGroupDevice.init_device(self)

        detect_evts = ()  # state and status are already set by the super class
        non_detect_evts = "configuration", "integrationtime", "monitorcount", \
                          "acquisitionmode", "elementlist"
        self.set_change_events(detect_evts, non_detect_evts)

        self.Elements = list(self.Elements)
        for i in range(len(self.Elements)):
            try:
                self.Elements[i] = int(self.Elements[i])
            except:
                pass
        mg = self.measurement_group
        if mg is None:
            full_name = self.get_full_name()
            name = self.alias or full_name
            self.measurement_group = mg = \
                self.pool.create_measurement_group(name=name,
                    full_name=full_name, id=self.Id,
                    user_elements=self.Elements)
        mg.add_listener(self.on_measurement_group_changed)

        # force a state read to initialize the state attribute
        #state = self.measurement_group.state
        self.set_state(DevState.ON)

    def on_measurement_group_changed(self, event_source, event_type, event_value):
        try:
            self._on_measurement_group_changed(event_source, event_type, event_value)
        except:
            msg = 'Error occured "on_measurement_group_changed(%s.%s): %s"'
            exc_info = sys.exc_info()
            self.error(msg, self.measurement_group.name, event_type.name,
                       exception_str(*exc_info[:2]))
            self.debug("Details", exc_info=exc_info)

    def _on_measurement_group_changed(self, event_source, event_type, event_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return

        timestamp = time.time()
        name = event_type.name
        name = name.replace('_', '')
        multi_attr = self.get_device_attr()
        attr = multi_attr.get_attr_by_name(name)
        quality = AttrQuality.ATTR_VALID
        priority = event_type.priority
        error = None

        if name == "state":
            event_value = self.calculate_tango_state(event_value)
        elif name == "status":
            event_value = self.calculate_tango_status(event_value)
        elif name == "acquisitionmode":
            event_value = AcqMode.whatis(event_value)
        elif name == "configuration":
            cfg = self.measurement_group.get_user_configuration()
            codec = CodecFactory().getCodec('json')
            _, event_value = codec.encode(('', cfg))
        else:
            if isinstance(event_value, SardanaAttribute):
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                timestamp = event_value.timestamp
                event_value = event_value.value

        self.set_attribute(attr, value=event_value, timestamp=timestamp,
                           quality=quality, priority=priority, error=error,
                           synch=False)

    def always_executed_hook(self):
        pass
        #state = to_tango_state(self.motor_group.get_state(cache=False))

    def read_attr_hardware(self, data):
        pass

    def read_IntegrationTime(self, attr):
        it = self.measurement_group.integration_time
        if it is None:
            it = float('nan')
        attr.set_value(it)

    def write_IntegrationTime(self, attr):
        self.measurement_group.integration_time = attr.get_write_value()

    def read_MonitorCount(self, attr):
        it = self.measurement_group.monitor_count
        if it is None:
            it = 0
        attr.set_value(it)

    def write_MonitorCount(self, attr):
        self.measurement_group.monitor_count = attr.get_write_value()

    def read_AcquisitionMode(self, attr):
        acq_mode = self.measurement_group.acquisition_mode
        acq_mode_str = AcqMode.whatis(acq_mode)
        attr.set_value(acq_mode_str)

    def write_AcquisitionMode(self, attr):
        acq_mode_str = attr.get_write_value()
        try:
            acq_mode = AcqMode.lookup[acq_mode_str]
        except KeyError:
            raise Exception("Invalid acquisition mode. Must be one of " + \
                            ", ".join(AcqMode.keys()))
        self.measurement_group.acquisition_mode = acq_mode

    def read_Configuration(self, attr):
        cfg = self.measurement_group.get_user_configuration()
        codec = CodecFactory().getCodec('json')
        data = codec.encode(('', cfg))
        attr.set_value(data[1])

    def write_Configuration(self, attr):
        data = attr.get_write_value()
        cfg = CodecFactory().decode(('json', data), ensure_ascii=True)
        self.measurement_group.set_configuration_from_user(cfg)

    def Start(self):
        try:
            self.wait_for_operation()
        except:
            raise Exception("Cannot acquire: already involved in an operation")
        self.measurement_group.start_acquisition()

    def StartMultiple(self, n):
        try:
            self.wait_for_operation()
        except:
            raise Exception("Cannot acquire: already involved in an operation")
        self.measurement_group.start_acquisition(multiple=n)

class MeasurementGroupClass(PoolGroupDeviceClass):

    #    Class Properties
    class_property_list = {
    }

    #    Device Properties
    device_property_list = {
    }
    device_property_list.update(PoolGroupDeviceClass.device_property_list)

    #    Command definitions
    cmd_list = {
        'Start': [ [DevVoid, ""], [DevVoid, ""] ],
        'StartMultiple': [ [DevLong, ""], [DevVoid, ""] ],
    }
    cmd_list.update(PoolGroupDeviceClass.cmd_list)

    #    Attribute definitions
    attr_list = {
        'IntegrationTime': [ [DevDouble, SCALAR, READ_WRITE],
                              { 'Memorized'     : "true",
                                'Display level' : DispLevel.OPERATOR } ],
        'MonitorCount': [ [DevLong, SCALAR, READ_WRITE],
                              { 'Memorized'     : "true",
                                'Display level' : DispLevel.OPERATOR } ],
        'AcquisitionMode': [ [DevString, SCALAR, READ_WRITE],
                              { 'Memorized'     : "true",
                                'Display level' : DispLevel.OPERATOR } ],
        'Configuration': [ [DevString, SCALAR, READ_WRITE],
                              { 'Memorized'     : "true",
                                'Display level' : DispLevel.EXPERT } ],
    }
    attr_list.update(PoolGroupDeviceClass.attr_list)

    def _get_class_properties(self):
        ret = PoolGroupDeviceClass._get_class_properties(self)
        ret['Description'] = "Measurement group device class"
        ret['InheritedFrom'].insert(0, 'PoolGroupDevice')
        return ret



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool libray. It defines the base classes
for external objects to the pool (like tango objects)"""

__all__ = ["PoolBaseExternalObject", "PoolTangoObject", "PoolExternalObject"]

__docformat__ = 'restructuredtext'

import PyTango

from sardana import ElementType
from sardana.pool.poolbaseobject import PoolBaseObject


class PoolBaseExternalObject(PoolBaseObject):
    """TODO"""

    def __init__(self, **kwargs):
        kwargs['elem_type'] = ElementType.External
        PoolBaseObject.__init__(self, **kwargs)

    def get_source(self):
        return self.full_name

    def get_config(self):
        raise NotImplementedError


class PoolTangoObject(PoolBaseExternalObject):
    """TODO"""

    def __init__(self, **kwargs):
        scheme = kwargs.pop('scheme', 'tango')
        attribute_name = kwargs.pop('attributename')
        host, port = kwargs.pop('host', None), kwargs.pop('port', None)
        devalias = kwargs.pop('devalias', None)
        device_name = kwargs.pop('devicename', None)
        if host is None:
            db = PyTango.Database()
            host, port = db.get_db_host(), db.get_db_port()
        else:
            db = PyTango.Database(host, port)
        full_name = "<unknown>"
        if device_name is None:
            if devalias is not None:
                try:
                    device_name = db.get_device_alias(devalias)
                    full_name = "{0}:{1}/{2}/{3}".format(host, port,
                                                         device_name,
                                                         attribute_name)
                except:
                    full_name = "{0}/{1}".format(devalias, attribute_name)
        else:
            full_name = "{0}:{1}/{2}/{3}".format(host, port, device_name,
                                                 attribute_name)
        self._device_name = device_name
        self._attribute_name = attribute_name
        self._config = None
        self._device = None
        kwargs['name'] = attribute_name
        kwargs['full_name'] = full_name
        PoolBaseExternalObject.__init__(self, **kwargs)

    def get_device_name(self):
        return self._device_name

    def get_attribute_name(self):
        return self._attribute_name

    def get_device(self):
        device = self._device
        if device is None:
            try:
                self._device = device = PyTango.DeviceProxy(self._device_name)
            except:
                pass
        return device

    def get_config(self):
        config = self._config
        if config is None:
            try:
                self._config = config = \
                    self._device.get_attribute_config(self._attribute_name)
            except:
                pass
        return config

    device_name = property(get_device_name)
    attribute_name = property(get_attribute_name)


_SCHEME_CLASS = { 'tango' : PoolTangoObject,
                  None    : PoolTangoObject}


def PoolExternalObject(**kwargs):
    scheme = kwargs.get('scheme', 'tango')
    klass = _SCHEME_CLASS[scheme]
    return klass(**kwargs)




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool libray. It defines the base classes
for pool event mechanism"""

from __future__ import absolute_import

__all__ = ["EventGenerator", "EventReceiver", "EventType"]

__docformat__ = 'restructuredtext'

import weakref
import collections

from sardana.sardanautils import is_callable
from taurus.core.util.event import CallableRef, BoundMethodWeakref


def _get_callable_ref(listener, callback=None):
    """Returns a callable reference for the given listener"""
    meth = getattr(listener, 'event_received', None)
    if meth is not None and is_callable(meth):
        return weakref.ref(listener, callback)
    else:
        return CallableRef(listener, callback)


class EventGenerator(object):
    """A class capable of generating events to their listeners"""

    def __init__(self, max_queue_len=10, listeners=None):
        self._listeners = []
        self._event_queue = collections.deque(maxlen=max_queue_len)
        if listeners is not None:
            if not isinstance(listeners, collections.Sequence):
                listeners = listeners,
            for listener in listeners:
                self.add_listener(listener)

    def _listener_died(self, weak_listener):
        """Callback executed when a listener dies"""
        if self._listeners is None:
            return
        try:
            self._listeners.remove(weak_listener)
        except ValueError:
            pass

    def add_listener(self, listener):
        """Adds a new listener for this object.
        
        :param listener: a listener"""
        if self._listeners is None or listener is None:
            return False

        weak_listener = _get_callable_ref(listener, self._listener_died)
        if weak_listener in self._listeners:
            return False
        self._listeners.append(weak_listener)
        return True

    def remove_listener(self, listener):
        """Removes an existing listener for this object.
        
        :param listener: the listener to be removed
        :return: True is succeeded or False otherwise"""
        if self._listeners is None:
            return
        weak_listener = _get_callable_ref(listener)
        try:
            self._listeners.remove(weak_listener)
        except ValueError:
            return False
        return True

    def has_listeners(self):
        """Returns True if anybody is listening to events from this object
        
        :return: True is at least one listener is listening or False otherwise
        """
        if self._listeners is None:
            return False
        return len(self._listeners) > 0

    def fire_event(self, event_type, event_value, listeners=None):
        self.flush_queue()
        self._fire_event(event_type, event_value, listeners=listeners)

    def _fire_event(self, event_type, event_value, listeners=None):
        """Sends an event to all listeners or a specific one"""
        if listeners is None:
            listeners = self._listeners
        if listeners is None:
            return
        if not isinstance(listeners, collections.Sequence):
            listeners = listeners,
        for listener in listeners:
            if isinstance(listener, weakref.ref) or \
               isinstance(listener, BoundMethodWeakref):
                real_listener = listener()
            else:
                real_listener = listener
            if real_listener is None:
                continue
            meth = getattr(real_listener, 'event_received', None)
            if meth is not None and is_callable(meth):
                real_listener.event_received(self, event_type, event_value)
            elif is_callable(real_listener):
                real_listener(self, event_type, event_value)

    def queue_event(self, event_type, event_value, listeners=None):
        queue = self._event_queue
        queue.append((event_type, event_value, listeners))

    def flush_queue(self):
        queue = self._event_queue
        n = len(queue)
        while n > 0:
            self._fire_event(*queue.pop())
            n = n - 1


class EventReceiver(object):
    """A simple class that implements useful features for a class which is 
    an event receiver. The actual class may inherit from this EventReceiver class
    and may choose to use just a part of the API provided by this class, the 
    whole API or none of the API."""

    def __init__(self):
        self._events_blocked = False

    def block_events(self):
        self._events_blocked = True

    def unblock_events(self):
        self._events_blocked = False

    def are_events_blocked(self):
        return self._events_blocked


class EventType(object):
    """Definition of an event type"""

    def __init__(self, name, priority=0):
        self.name = name
        self.priority = priority

    def __str__(self):
        return "EventType(name=%s, priority=%s)" % (self.name, self.priority)

    def __repr__(self):
        return "EventType(name=%s, priority=%s)" % (self.name, self.priority)

    def get_name(self):
        """Returns this event name
        
        :return: this event name
        :rtype: str"""
        return self.name

    def get_priority(self):
        """Returns this event priority
        
        :return: this event priority
        :rtype: str"""
        return self.priority



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains a taurus text editor widget."""

__all__ = ["ControllerClassTreeWidget", "ControllerClassSelectionDialog"]

__docformat__ = 'restructuredtext'

import sys
import os

import taurus.core
from taurus.core.util.enumeration import Enumeration
from taurus.external.qt import Qt
from taurus.qt.qtcore.mimetypes import TAURUS_MODEL_MIME_TYPE, TAURUS_MODEL_LIST_MIME_TYPE
from taurus.qt.qtcore.model import TaurusBaseTreeItem, TaurusBaseModel, TaurusBaseProxyModel
from taurus.qt.qtgui.tree import TaurusBaseTreeWidget
from taurus.qt.qtgui.resource import getThemeIcon, getIcon

PoolControllerView = Enumeration("PoolControllerView", ("ControllerModule", "ControllerClass", "Unknown"))

def getElementTypeIcon(t):
    if t == PoolControllerView.ControllerModule:
        return getIcon(":/python-file.png")
    elif t == PoolControllerView.ControllerClass:
        return getIcon(":/python.png")
    return getIcon(":/tango.png")

def getElementTypeSize(t):
    return Qt.QSize(200, 24)

def getElementTypeToolTip(t):
    """Wrapper to prevent loading qtgui when this module is imported"""
    if t == PoolControllerView.ControllerModule:
        return "Controller module"
    elif t == PoolControllerView.ControllerClass:
        return "Controller class"


class ControllerBaseTreeItem(TaurusBaseTreeItem):
    """A generic node"""

    def data(self, index):
        """Returns the data of this node for the given index
        
        :return: (object) the data for the given index
        """
        return self._itemData

    def role(self):
        """Returns the prefered role for the item.
        This implementation returns taurus.core.taurusbasetypes.TaurusElementType.Unknown
        
        This method should be able to return any kind of python object as long
        as the model that is used is compatible.
        
        :return: (PoolControllerView) the role in form of element type"""
        return PoolControllerView.Unknown


class ControllerModuleTreeItem(ControllerBaseTreeItem):

    def role(self):
        return PoolControllerView.ControllerModule

    def toolTip(self):
        return "The controller module '%s'" % self.display()

    def icon(self):
        return getIcon(":/python-file.png")


class ControllerTreeItem(ControllerBaseTreeItem):

    def data(self, index):
        """Returns the data of this node for the given index
        
        :return: (object) the data for the given index
        """
        return self._itemData.name

    def role(self):
        return PoolControllerView.ControllerClass

    def toolTip(self):
        return self._itemData.doc

    def icon(self):
        return getIcon(":/python.png")


class ControllerBaseModel(TaurusBaseModel):

    ColumnNames = "Controllers",
    ColumnRoles = (PoolControllerView.ControllerModule, PoolControllerView.ControllerModule, PoolControllerView.ControllerClass),

    def setDataSource(self, pool):
        if self._data_src is not None:
            Qt.QObject.disconnect(self._data_src, Qt.SIGNAL('controllerClassesUpdated'), self.controllerClassesUpdated)
        if pool is not None:
            Qt.QObject.connect(pool, Qt.SIGNAL('controllerClassesUpdated'), self.controllerClassesUpdated)
        TaurusBaseModel.setDataSource(self, pool)

    def controllerClassesUpdated(self):
        self.refresh()

    def createNewRootItem(self):
        return ControllerBaseTreeItem(self, self.ColumnNames)

    def roleIcon(self, role):
        return getElementTypeIcon(role)

    def columnIcon(self, column):
        return self.roleIcon(self.role(column))

    def roleToolTip(self, role):
        return getElementTypeToolTip(role)

    def columnToolTip(self, column):
        return self.roleToolTip(self.role(column))

    def roleSize(self, role):
        return getElementTypeSize(role)

    def columnSize(self, column):
        role = self.role(column)
        s = self.roleSize(role)
        return s

    def mimeTypes(self):
        return ["text/plain", TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_MODEL_MIME_TYPE]

    def mimeData(self, indexes):
        ret = Qt.QMimeData()
        data = []
        for index in indexes:
            if not index.isValid(): continue
            tree_item = index.internalPointer()
            mime_data_item = tree_item.mimeData(index)
            if mime_data_item is None:
                continue
            data.append(mime_data_item)
        ret.setData(TAURUS_MODEL_LIST_MIME_TYPE, "\r\n".join(data))
        ret.setText(", ".join(data))
        if len(data) == 1:
            ret.setData(TAURUS_MODEL_MIME_TYPE, str(data[0]))
        return ret

    def pyData(self, index, role):
        if not index.isValid():
            return None

        item = index.internalPointer()
        row, column, depth = index.row(), index.column(), item.depth()
        item_role = self.role(column, depth)

        ret = None
        if role == Qt.Qt.DisplayRole:
            ret = Qt.QString(item.data(index))
        elif role == Qt.Qt.DecorationRole:
            ret = item.icon()
        elif role == Qt.Qt.ToolTipRole:
            ret = item.toolTip()
        #elif role == Qt.Qt.SizeHintRole:
        #    ret = self.columnSize(column)
        elif role == Qt.Qt.FontRole:
            ret = self.DftFont
        return ret

    def setupModelData(self, data):
        pool = self.dataSource()
        if pool is None:
            return
        root = self._rootItem
        ctrl_modules = {}
        # TODO
        #ctrl_class_dict = pool.getControllerClasses()
        #for ctrl_class_name, ctrl_class in ctrl_class_dict.items():
        #    module_name = ctrl_class.module_name
        #    moduleNode = ctrl_modules.get(module_name)
        #    if moduleNode is None:
        #        moduleNode = ControllerModuleTreeItem(self, module_name, root)
        #        root.appendChild(moduleNode)
        #        ctrl_modules[module_name] = moduleNode
        #    ctrlNode = ControllerTreeItem(self, ctrl_class, moduleNode)
        #    moduleNode.appendChild(ctrlNode)


class ControllerModuleModel(ControllerBaseModel):
    pass


class PlainControllerModel(ControllerBaseModel):

    ColumnNames = "Controller classes",
    ColumnRoles = (PoolControllerView.ControllerClass, PoolControllerView.ControllerClass),

    def setupModelData(self, data):
        pool = self.dataSource()
        if pool is None:
            return
        root = self._rootItem
        # TODO
        #ctrl_class_dict = pool.getControllerClasses()
        #ctrl_classes = ctrl_class_dict.keys()
        #ctrl_classes.sort()
        #self.debug("Found %d controller classes", len(ctrl_classes))
        #for ctrl_class_name in ctrl_classes:
        #    ctrl_class = ctrl_class_dict[ctrl_class_name]
        #    ctrlNode = ControllerTreeItem(self, ctrl_class, root)
        #    root.appendChild(ctrlNode)


class ControllerBaseModelProxy(TaurusBaseProxyModel):
    pass


class ControllerModuleModelProxy(ControllerBaseModelProxy):
    pass


class PlainControllerModelProxy(ControllerBaseModelProxy):
    pass


class ControllerClassTreeWidget(TaurusBaseTreeWidget):

    KnownPerspectives = { PoolControllerView.ControllerModule : {
                            "label" : "By module",
                            "icon" : ":/python-file.png",
                            "tooltip" : "View by controller module",
                            "model" : [ControllerModuleModelProxy, ControllerModuleModel],
                          },
                          PoolControllerView.ControllerClass : {
                            "label" : "By controller",
                            "icon" : ":/python.png",
                            "tooltip" : "View by controller class",
                            "model" : [PlainControllerModelProxy, PlainControllerModel],
                          }
    }
    DftPerspective = PoolControllerView.ControllerModule

    def getModelClass(self):
        return taurus.core.taurusdevice.TaurusDevice


class ControllerClassSelectionDialog(Qt.QDialog):

    def __init__(self, parent=None, designMode=False, model_name=None, perspective=None):
        Qt.QDialog.__init__(self, parent)

        self.setWindowTitle("Controller Class Selection Dialog")

        layout = Qt.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self._panel = ControllerClassTreeWidget(parent=self,
                                                perspective=perspective,
                                                designMode=designMode,
                                                with_navigation_bar=False)
        self._panel.setModel(model_name)
        self.setWindowIcon(getElementTypeIcon(self._panel.perspective()))
        self._buttonBox = Qt.QDialogButtonBox(self)
        bts = Qt.QDialogButtonBox.Ok | Qt.QDialogButtonBox.Cancel
        self._buttonBox.setStandardButtons(bts)
        layout.addWidget(self._panel)
        layout.addWidget(self._buttonBox)
        self.connect(self._buttonBox, Qt.SIGNAL("accepted()"), self.accept)
        self.connect(self._buttonBox, Qt.SIGNAL("rejected()"), self.reject)

    def selectedItems(self):
        return self._panel.selectedItems()

    def getSelectedMacros(self):
        return [ i.itemData() for i in self.selectedItems() ]


def main_ControllerClassSelecionDialog(pool, perspective=PoolControllerView.ControllerClass):
    w = ControllerClassSelectionDialog(model_name=pool, perspective=perspective)

    if w.result() == Qt.QDialog.Accepted:
        print w.getSelectedMacros()
    return w

def main_ControllerClassTreeWidget(pool, perspective=PoolControllerView.ControllerClass):
    w = ControllerClassTreeWidget(perspective=perspective, with_navigation_bar=False)
    w.setModel(pool)
    w.show()
    return w

def demo(poolname="Pool_BL98"):
    """ControllerClassTreeWidget"""
    w = main_ControllerClassSelecionDialog(poolname, PoolControllerView.ControllerClass)
    return w

def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        app = Application(app_name="Pool controller class tree demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    args = app.get_command_line_args()
    if len(args) == 1:
        w = demo(poolname=args[0])
    else:
        w = demo()

    w.show()
    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == "__main__":
    main()



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

from taurus.external.qt import Qt
from PyQt4 import Qsci

from taurus.qt.qtgui.resource import getThemeIcon


class MacroEditor(Qsci.QsciScintilla):
    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent=None, designMode=False):
        Qsci.QsciScintilla.__init__(self, parent)

        self.textEdit = Qsci.QsciScintilla()
        self.textEdit.setAutoCompletionThreshold(1)
        self.textEdit.setAutoCompletionSource(Qsci.QsciScintilla.AcsAll)
        self.textEdit.setAutoIndent(True)
        self.textEdit.setCallTipsStyle(Qsci.QsciScintilla.CallTipsContext)
        self.textEdit.setCallTipsVisible(0)
        self.pythonLexer = Qsci.QsciLexerPython(self.textEdit)
        self.api = Qsci.QsciAPIs(self.pythonLexer)
        self.api.add(Qt.QString("dupa(a,b)this is function dupa"))
        self.api.prepare()
        self.pythonLexer.setAPIs(self.api)
        self.textEdit.setLexer(self.pythonLexer)

        self.newAction = Qt.QAction(getThemeIcon("document-new"), "New", self)
        self.connect(self.newAction, Qt.SIGNAL("triggered()"), self.newFile)
        self.newAction.setToolTip("Create new file")
        self.newAction.setShortcut("Ctrl+N")

        self.openAction = Qt.QAction(getThemeIcon("document-open"), "Open", self)
        self.connect(self.openAction, Qt.SIGNAL("triggered()"), self.openFile)
        self.openAction.setToolTip("Open existing file")
        self.openAction.setShortcut("Ctrl+O")

        self.saveAction = Qt.QAction(getThemeIcon("document-save"), "Save", self)
        self.connect(self.saveAction, Qt.SIGNAL("triggered()"), self.saveFile)
        self.saveAction.setToolTip("Save document to disk")
        self.saveAction.setShortcut("Ctrl+S")

        self.saveAsAction = Qt.QAction(getThemeIcon("document-save-as"), "Save as...", self)
        self.connect(self.saveAsAction, Qt.SIGNAL("triggered()"), self.saveFile)
        self.saveAsAction.setToolTip("Save document under a new name")

        self.cutAction = Qt.QAction(getThemeIcon("edit-cut"), "Cut", self)
        self.connect(self.cutAction, Qt.SIGNAL("triggered()"), self.cut)
        self.cutAction.setToolTip("Cut current selection's contents to the clipboard")
        self.cutAction.setShortcut("Ctrl+X")
        self.cutAction.setEnabled(False)

        self.copyAction = Qt.QAction(getThemeIcon("edit-copy"), "Copy", self)
        self.connect(self.copyAction, Qt.SIGNAL("triggered()"), self.copy)
        self.copyAction.setToolTip("Copy current selection's contents to the clipboard")
        self.copyAction.setShortcut("Ctrl+C")
        self.copyAction.setEnabled(False)

        self.pasteAction = Qt.QAction(getThemeIcon("edit-paste"), "Paste", self)
        self.connect(self.pasteAction, Qt.SIGNAL("triggered()"), self.paste)
        self.pasteAction.setToolTip("Paste the clipboard's contents into the current selection")
        self.pasteAction.setShortcut("Ctrl+V")

        self.aboutAction = Qt.QAction("About", self)
        self.connect(self.aboutAction, Qt.SIGNAL("triggered()"), self.about)
        self.aboutAction.setToolTip("Show the application's About box")

        self.connect(self.textEdit, Qt.SIGNAL("copyAvailable(bool)"), self.cutAction.setEnabled)
        self.connect(self.textEdit, Qt.SIGNAL("copyAvailable(bool)"), self.copyAction.setEnabled)

        self.setCurrentFile("")

    def closeEvent(self, event):
        if self.maybeSave():
#            self.writeSettings()
            event.accept()
        else:
            event.ignore()

    def newFile(self):
        if self.maybeSave():
            self.textEdit.clear()
            self.setCurrentFile("")

    def openFile(self):
        if self.maybeSave():
            fileName = Qt.QFileDialog.getOpenFileName(self)
        if not fileName is None and file != "":
            self.loadFile(fileName)

    def saveFile(self):
        if self.curFile == "":
            return self.__saveAs()
        else:
            return self.__saveFile(self.curFile)

    def __saveAs(self):
        self.fileName = Qt.QFileDialog.getSaveFileName(self)
        if self.fileName == "":
            return False
        return self.__saveFile(self.fileName)

    def about(self):
        Qt.QMessageBox.about(self, "About MacroEditor", "The MacroEditor by Zbigniew Reszela")

    def documentWasModified(self):
        self.setWindowModified(self.textEdit.isModified())

    def createMenus(self):
        fileMenu = self.menuBar().addMenu("File")
        fileMenu.addAction(self.newAction)
        fileMenu.addAction(self.openAction)
        fileMenu.addAction(self.saveAction)
        fileMenu.addAction(self.saveAsAction)

        editMenu = self.menuBar().addMenu("Edit")
        editMenu.addAction(self.cutAction)
        editMenu.addAction(self.copyAction)
        editMenu.addAction(self.pasteAction)

        self.menuBar().addSeparator()

        helpMenu = self.menuBar().addMenu("&Help")
        helpMenu.addAction(self.aboutAction)

#    def createToolBars(self):
#        fileToolBar = self.addToolBar("File")
#        fileToolBar.setIconSize(Qt.QSize(36,36))
#        fileToolBar.addAction(self.newAction)
#        fileToolBar.addAction(self.openAction)
#        fileToolBar.addAction(self.saveAction)
#        fileToolBar.addAction(self.saveAsAction)
#
#        editToolBar = self.addToolBar("Edit")
#        editToolBar.setIconSize(Qt.QSize(36,36))
#        editToolBar.addAction(self.cutAction)
#        editToolBar.addAction(self.copyAction)
#        editToolBar.addAction(self.pasteAction)

    def createStatusBar(self):
        self.statusBar().showMessage("Ready")

    def maybeSave(self):
        if self.textEdit.isModified():
            ret = Qt.QMessageBox.warning(self, "MacroEditor",
                     "The document has been modified\nDo you want to save your changes?",
                     Qt.QMessageBox.Yes | Qt.QMessageBox.Default,
                     Qt.QMessageBox.No,
                     Qt.QMessageBox.Cancel | Qt.QMessageBox.Escape)
            if ret == Qt.QMessageBox.Yes:
                return self.save()
            elif ret == Qt.QMessageBox.Cancel:
                return False
        return True

    def loadFile(self, fileName):
        try:
            fileHandle = open(fileName, 'r')
        except IOError, e:
            Qt.QMessageBox.warning(self, "MacroEditor", "Cannot read file %s:\n%s." % (fileName, e))
            return False
        fileContents = fileHandle.read()
        Qt.QApplication.setOverrideCursor(Qt.Qt.WaitCursor)
        self.textEdit.setText(fileContents)
        Qt.QApplication.restoreOverrideCursor()

        self.setCurrentFile(fileName)
        self.statusBar().showMessage("File loaded", 2000)

    def __saveFile(self, fileName):
        try:
            file = open(fileName, 'w')
        except IOError, e:
            Qt.QMessageBox.warning(self, "MacroEditor", "Cannot write file %s:\n%s." % (fileName, e))
            return False
        Qt.QApplication.setOverrideCursor(Qt.Qt.WaitCursor)
        file.write(self.textEdit.text())
        Qt.QApplication.restoreOverrideCursor()
        self.setCurrentFile(fileName)
        self.statusBar().showMessage(("File saved"), 2000)
        return True

    def setCurrentFile(self, fileName):
        self.curFile = fileName
        self.textEdit.setModified(False)
        self.setWindowModified(False)

        if self.curFile == "":
            shownName = "untitled.txt"
        else:
            shownName = self.strippedName(self.curFile)
        self.setWindowTitle("%s[*] - %s" % (shownName, "MacroExecutor"))

    def strippedName(self, fullFileName):
        return Qt.QFileInfo(fullFileName).fileName()

    def cut(self):
        pass

    def copy(self):
        pass

    def paste(self):
        pass


def test():
    import sys
    app = Qt.QApplication(sys.argv)
    macroEditor = MacroEditor()
    macroEditor.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    test()



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for"""

__all__ = ["PoolBaseGroup"]

__docformat__ = 'restructuredtext'

from taurus.core.taurusvalidator import AttributeNameValidator

from sardana import State, ElementType, TYPE_PHYSICAL_ELEMENTS
from sardana.pool.poolexternal import PoolExternalObject
from sardana.pool.poolcontainer import PoolContainer


class PoolBaseGroup(PoolContainer):

    def __init__(self, **kwargs):
        self._pending = True
        self._user_element_ids = None
        self._user_elements = None
        self._physical_elements = None
        self._physical_elements_set = None
        self._state_statistics = {}
        self._pool = kwargs.pop('pool')
        self.set_user_element_ids(kwargs.pop('user_elements'))
        PoolContainer.__init__(self)
        try:
            self._build_elements()
        except KeyError:
            self.info("failed to build element information. No problem. " \
                      "Probably one or more underlying elements have not " \
                      "been constructed yet")

    def _get_pool(self):
        return self._pool

    def _create_action_cache(self):
        raise NotImplementedError

    def _get_action_cache(self):
        if self._action_cache is None:
            self._action_cache = self._fill_action_cache()
        return self._action_cache

    def _set_action_cache(self, action_cache):
        physical_elements = self.get_physical_elements()
        if self._action_cache is not None:
            for ctrl_physical_elements in physical_elements.values():
                for physical_element in ctrl_physical_elements:
                    action_cache.remove_element(physical_element)

        self._action_cache = self._fill_action_cache(action_cache)

    def _fill_action_cache(self, action_cache=None, physical_elements=None):
        if action_cache is None:
            action_cache = self._create_action_cache()
        if physical_elements is None:
            physical_elements = self.get_physical_elements()
        for _, ctrl_physical_elements in physical_elements.items():
            for physical_element in ctrl_physical_elements:
                action_cache.add_element(physical_element)
        return action_cache

    def _calculate_element_state(self, elem, elem_state_info):
        u_state, u_status = elem_state_info
        if u_status is None:
            u_status = '%s is None' % elem.name
        else:
            u_status = u_status.split("\n", 1)[0]
        return u_state, u_status

    def _calculate_states(self, state_info=None):
        user_elements = self.get_user_elements()
        none, unknown = set(), set()
        fault, alarm, on, moving = set(), set(), set(), set()
        status = []
        if state_info is None:
            state_info = {}
            for elem in user_elements:
                if elem.get_type() == ElementType.External:
                    continue
                # cannot call get_state(us) here since it may lead to dead lock!
                si = elem.inspect_state(), elem.inspect_status()
                state_info[elem] = si
        for elem, elem_state_info in state_info.items():
            elem_type = elem.get_type()
            if elem_type == ElementType.External:
                continue
            u_state, u_status = self._calculate_element_state(elem, elem_state_info)
            if u_state == State.Moving:
                moving.add(elem)
            elif u_state == State.On:
                on.add(elem)
            elif u_state == State.Fault:
                fault.add(elem)
            elif u_state == State.Alarm:
                alarm.add(elem)
            elif u_state is State.Unknown:
                unknown.add(elem)
            elif u_state is None:
                none.add(elem)
            status.append(u_status)
        state = State.On
        if none or unknown:
            state = State.Unknown
        if fault:
            state = State.Fault
        elif alarm:
            state = State.Alarm
        elif moving:
            state = State.Moving
        self._state_statistics = { State.On : on, State.Fault : fault,
                                   State.Alarm : alarm, State.Moving : moving,
                                   State.Unknown : unknown, None : none }
        status = "\n".join(status)
        return state, status

    def _is_managed_element(self, element):
        return True

    def _build_elements(self):
        self._user_elements = []
        self._physical_elements = {}
        self._physical_elements_set = set()

        pool = self._get_pool()
        for user_element_id in self._user_element_ids:
            # an internal element
            internal = type(user_element_id) is int
            if internal:
                try:
                    user_element = pool.get_element(id=user_element_id)
                except KeyError:
                    self._pending = True
                    self._user_elements = None

                    self._physical_elements = None
                    self._physical_elements_set = None
                    raise
                internal = self._is_managed_element(user_element)
                if not internal:
                    user_element_id = user_element.get_source()
            # a tango channel or non internal element (ex: ioregister or motor
            # in measurement group)
            if not internal:
                validator = AttributeNameValidator()
                params = validator.getParams(user_element_id)
                params['pool'] = self._get_pool()
                user_element = PoolExternalObject(**params)
            self.add_user_element(user_element)
        self._pending = False

    def on_element_changed(self, evt_src, evt_type, evt_value):
        pass

    def set_user_element_ids(self, new_element_ids):
        self.clear_user_elements()
        self._user_element_ids = new_element_ids

    def get_user_element_ids(self):
        """Returns the sequence of user element IDs
        
        :return: the sequence of user element IDs
        :rtype: sequence< :obj:`int`>"""
        return self._user_element_ids

    user_element_ids = property(get_user_element_ids)

    def get_user_elements(self):
        """Returns the sequence of user elements
        
        :return: the sequence of user elements
        :rtype: sequence< :class:`~sardana.pool.poolelement.PoolElement`>"""
        if self._pending:
            self._build_elements()
        return self._user_elements

    def get_user_elements_attribute_iterator(self):
        """Returns an iterator over the main attribute of each user element.
        
        :return: an iterator over the main attribute of each user element.
        :rtype: iter< :class:`~sardana.sardanaattribute.SardanaAttribute` >"""
        for element in self.get_user_elements():
            yield element.get_default_attribute()

    get_user_elements_attribute = get_user_elements_attribute_iterator
    get_user_elements_attribute.__doc__ = get_user_elements_attribute_iterator.__doc__

    def get_user_elements_attribute_sequence(self):
        """Returns a sequence of main attribute of each user element.
        
        In loops use preferably :meth:`get_user_elements_attribute_iterator` for
        performance and memory reasons.
        
        :return: a sequence of main attribute of each user element.
        :rtype: sequence< :class:`~sardana.sardanaattribute.SardanaAttribute` >"""
        return list(self.get_user_elements_attribute_iterator())

    def get_user_elements_attribute_map(self):
        """Returns a dictionary of main attribute of each user element.
        
        :return: a dictionary of main attribute of each user element.
        :rtype: dict< :class:`~sardana.pool.poolelement.PoolElement`, 
                :class:`~sardana.sardanaattribute.SardanaAttribute` >"""
        ret = {}
        for element in self.get_user_elements():
            ret[element] = element.get_default_attribute()
        return ret

    def get_physical_elements(self):
        """Returns a dictionary or physical elements where key is a controller
        object and value is a sequence of pool elements
        
        :return: a dictionary of physical elements
        :rtype: dict< :class:`~sardana.pool.poolcontroller.PoolController,
                sequence<:class:`~sardana.pool.poolelement.PoolElement`>"""
        if self._pending:
            self._build_elements()
        return self._physical_elements

    def get_physical_elements_iterator(self):
        """Returns an iterator over the physical elements.
        
        .. warning:: The order is non deterministic.
        
        :return: an iterator over the physical elements.
        :rtype: iter<:class:`~sardana.pool.poolelement.PoolElement` >"""
        for _, elements in self.get_physical_elements().items():
            for element in elements:
                yield element

    def get_physical_elements_attribute_iterator(self):
        """Returns an iterator over the main attribute of each physical element.
        
        .. warning:: The order is non deterministic.
        
        :return: an iterator over the main attribute of each physical element.
        :rtype: iter< :class:`~sardana.sardanaattribute.SardanaAttribute` >"""
        for element in self.get_physical_elements_iterator():
            yield element.get_default_attribute()

    def get_physical_elements_set(self):
        if self._pending:
            self._build_elements()
        return self._physical_elements_set

    def add_user_element(self, element, index=None):
        user_elements = self._user_elements
        physical_elements = self._physical_elements
        physical_elements_set = self._physical_elements_set

        if element in user_elements:
            raise Exception("Group already contains %s" % element.name)
        if index is None:
            index = len(user_elements)
        user_elements.insert(index, element)

        if not self._is_managed_element(element):
            return index

        self.add_element(element)
        self._find_physical_elements(element,
                                     physical_elements=physical_elements,
                                     physical_elements_set=physical_elements_set)
        action_cache = self._action_cache
        if action_cache is not None:
            self._fill_action_cache(action_cache=action_cache,
                                    physical_elements=physical_elements)
        element.add_listener(self.on_element_changed)
        return index

    def _find_physical_elements(self, element, physical_elements=None,
                                physical_elements_set=None):
        elem_type = element.get_type()
        if physical_elements is None:
            physical_elements = {}
        if physical_elements_set is None:
            physical_elements_set = set()

        if elem_type in TYPE_PHYSICAL_ELEMENTS:
            ctrl = element.controller
            own_elements = physical_elements.get(ctrl)
            if own_elements is None:
                physical_elements[ctrl] = own_elements = set()
            own_elements.add(element)
            physical_elements_set.add(element)
        else:
            for ctrl, elements in element.get_physical_elements().items():
                own_elements = physical_elements.get(ctrl)
                if own_elements is None:
                    physical_elements[ctrl] = own_elements = set()
                own_elements.update(elements)
                physical_elements_set.update(elements)
        return physical_elements

    # TODO: too complicated to implement for now
#    def remove_user_element(self, element):
#        try:
#            idx = self._user_elements.index(element)
#        except ValueError:
#            raise Exception("Group doesn't contain %s" % element.name)
#        action_cache = self.get_action_cache()
#        element.remove_listener(self.on_element_changed)
#        action_cache.remove_element(element)
#        del self._user_elements[idx]
#        del self._user_element_ids[self._user_element_ids.index(element.id)]
#        self.remove_element(element)

    def clear_user_elements(self):
        user_elements = self._user_elements
        if user_elements is not None:
            for element in user_elements:
                if element.get_type() != ElementType.External:
                    element.remove_listener(self.on_element_changed)
                    self.remove_element(element)
        self._action_cache = None
        self._pending = True
        self._user_elements = None
        self._user_element_ids = None
        self._physical_elements = None


    # --------------------------------------------------------------------------
    # stop
    # --------------------------------------------------------------------------

    def stop(self):
        for ctrl, elements in self.get_physical_elements().items():
            self.debug("Stopping %s %s", ctrl.name, [e.name for e in elements])
            try:
                ctrl.stop_elements(elements=elements)
            except:
                self.error("Unable to stop controller %s", ctrl.name)
                self.debug("Details:", exc_info=1)

    # --------------------------------------------------------------------------
    # abort
    # --------------------------------------------------------------------------

    def abort(self):
        for ctrl, elements in self.get_physical_elements().items():
            self.debug("Aborting %s %s", ctrl.name, [e.name for e in elements])
            try:
                ctrl.abort_elements(elements=elements)
            except:
                self.error("Unable to abort controller %s", ctrl.name)
                self.debug("Details:", exc_info=1)

    # --------------------------------------------------------------------------
    # involved in an operation
    # --------------------------------------------------------------------------

    def get_operation(self):
        for _, elements in self.get_physical_elements().items():
            for element in elements:
                op = element.get_operation()
                if op is not None:
                    return op
        return None



#!/usr/bin/env python

#############################################################################
##
## This file is part of Taurus, a Tango User Interface Library
## 
## http://www.tango-controls.org/static/taurus/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains some Sardana-wide default configurations.

The idea is that the final user may edit the values here to customize certain
aspects of Sardana.
"""


#:UnitTets door name: is the door to use in our unit tests.
#:Its definition is a mandatory for the unittests. 
#:In case the Door is not valid or defined an exception will be raise.
UNITTEST_DOOR_NAME = "door/demo1/1"

#:Variable controlling the size and number of rotating backups of the log files.
#:The Pool and MacroServer Device servers will use these values for their logs. 
LOG_FILES_SIZE = 1e7
LOG_BCK_COUNT = 5